/* File *******************************************************************
*					North Carolina State University
*					Electronics Research Laboratory
*
* FILENAME :    s2ibis.lex
*
* DESCRIPTION : This file provides the lexical analyzer which will pass 
*				tokens to the parser.  It implements the command language
*				described in the s2ibis2 documentation.
*
*
* NOTES :		While the command language is very similar to the IBIS 
*				command language, this file will require a bit of revision
*				to correctly tokenize IBIS commands.
*
*
* REVISION HISTORY :
*       Date        Programmer:         Description
*		8/24/95		Alan Glaser			Beta release to some members of the
*					Michael Steer		IBIS community
*
*
*H*/

/************************************************************************
Note: the following lines go between the lines

    YY_FATAL_ERROR( "read() in flex scanner failed" ); \

	and 

  for(i = result, b = buf; i ; i--)\

in #define YY_INPUT below.

#ifdef LEX_DEBUG
  printf("result = %d \n",result);\
  printf("buf    =\n");\
  for(i = result, b = buf; i ; i--) putchar(*(b++));;\
  printf("\n");\ 
#endif


************************************************************************/
	  

%{
/*
 * Redefine YY_INPUT so that one character at a time is read.  This
 * facilitates skipping over comments which begin with a comment character
 * which can be changed on the fly.
 */
#undef YY_INPUT
#define YY_INPUT(buf,result,max_size) \
  { \
  int c = getc(yyin); \
/* skip comments beginning with the comment character up to EOL or EOF */ \
  if(c == commentChar) \
    while((c = getc(yyin)) != '\n' && c != EOF); \
  result = (c == EOF) ? YY_NULL : (buf[0] = c, 1); \
  }
#undef YY_INPUT
#define YY_INPUT(buf,result,max_size) \
  { \
  int i; char *b;\
  if ( (result = read( fileno(yyin), (char *) buf, max_size )) < 0 ) \
    YY_FATAL_ERROR( "read() in flex scanner failed" ); \
  for(i = result, b = buf; i ; i--)\
    if(*(b++) == commentChar) *(b-1) = universalCommentIndicator;\
  }


int yywrap()
  {
  return(1);
  }
%}

%x S_PIN S_COMPONENT S_MODEL S_MODEL_TYPE S_GET_VERSION S_SPICE_TYPE 

/*
 * lf is linefeed   or   verticle tab  or  form feed  or carriage return
 */
white			[\t" "]
predigit		("-"|"+")?
lf				[\x0d]*[\n]
digits			([0-9])+
expon			[Ee](["-""+""^"]?)([0-9]+)
float_core		(([0-9])*\.([0-9])+)|(([0-9])+\.([0-9])*)
not_eol_eof_white	[\x00-\x08\x0e-\x1f\x21-\xfd]
comment			[\x7f]
empty			[]

%%


<INITIAL>^"["IBIS{white}+Ver"]"	{ 
	BEGIN( S_GET_VERSION );
	RET(IBIS_VERSION) 
	}

<INITIAL>^"["File{white}+name"]"		{ RET( FILE_NAME ) }

<INITIAL>^"["File{white}+rev"]"			{ RET( FILE_REV ) }

<INITIAL>^"["Iterate"]"					{ RET( ITERATE ) }

<INITIAL>^"["Cleanup"]"					{ RET( CLEANUP ) }

<INITIAL>^"["Date"]"					{ RET( DATE ) }

<INITIAL>^"["Source"]"					{ RET( SOURCE ) }

<INITIAL>^"["Notes"]"					{ RET( NOTES ) }

<INITIAL>^"["Disclaimer"]"				{ RET( DISCLAIMER ) }

<INITIAL>^"["Copyright"]"				{ RET( COPYRIGHT ) }

<INITIAL>^"["Temperature{white}+range"]" { RET( GLOBAL_TEMP_RANGE ) }

<INITIAL>^"["Voltage{white}+range"]"	{ RET( GLOBAL_VOLTAGE_RANGE ) }

<INITIAL>^"["Pullup{white}+reference"]"	{ RET( GLOBAL_PULLUP ) }

<INITIAL>^"["Pulldown{white}+reference"]" { RET( GLOBAL_PULLDOWN ) }

<INITIAL>^"["POWER{white}+clamp{white}+reference"]"	{ RET( GLOBAL_POWER_CLAMP ) }

<INITIAL>^"["GND{white}+clamp{white}+reference"]"	{ RET( GLOBAL_GND_CLAMP ) }

<INITIAL>^"["Rload"]"					{ RET( GLOBAL_RLOAD ) }

<INITIAL>^"["Derate{white}+VI"]"		{ RET( GLOBAL_DERATE_VI ) }

<INITIAL>^"["Derate{white}+ramp"]"		{ RET( GLOBAL_DERATE_RAMP ) }

<INITIAL>^"["R_pkg"]"					{ RET( GLOBAL_R_PKG ) }

<INITIAL>^"["L_pkg"]"					{ RET( GLOBAL_L_PKG ) }

<INITIAL>^"["C_pkg"]"					{ RET( GLOBAL_C_PKG ) }

<INITIAL>^"["C_comp"]"					{ RET( GLOBAL_C_COMP ) }

<INITIAL>^"["Sim{white}+time"]"			{ RET( GLOBAL_SIM_TIME ) }

<INITIAL>^"["Spice{white}+command"]"	{ RET( SPICE_COMMAND ) }

<INITIAL>^"["Spice{white}+type"]"		{ 
	BEGIN( S_SPICE_TYPE );
	RET( SPICE_TYPE ) 
	}



<*>^"["Diff{white}+Pin"]"	{ 
	RET( DIFF_PIN ) 
	}

<*>^"["Model"]"		{ 
	BEGIN( S_MODEL );
	RET( MODEL )
	}

<*>^"["Pin"]"		{ 
	BEGIN( S_PIN );
	RET( PIN ) 
	}

<*>^"["Pin{white}+Mapping"]"	{ 
	RET( PIN_MAP ) 
	}

<*>^"["Component"]"	{ 
	BEGIN( S_COMPONENT );
	RET( COMPONENT ) 
	}



<S_COMPONENT>^"["Manufacturer"]"			{ RET( MANUFACTURER ) }

<S_COMPONENT>^"["Package{white}+model"]"	{ RET( PACKAGE_MODEL ) }

<S_COMPONENT>^"["Spice{white}+file"]"		{ RET( SPICE_FILE ) }

<S_COMPONENT>^"["Temperature{white}+range"]" { RET( COMPONENT_TEMP_RANGE ) }

<S_COMPONENT>^"["Voltage{white}+range"]"	{ RET( COMPONENT_VOLTAGE_RANGE ) }

<S_COMPONENT>^"["Pullup{white}+reference"]"	{ RET( COMPONENT_PULLUP ) }

<S_COMPONENT>^"["Pulldown{white}+reference"]" { RET( COMPONENT_PULLDOWN ) }

<S_COMPONENT>^"["POWER{white}+clamp{white}+reference"]"	{ RET( COMPONENT_POWER_CLAMP ) }

<S_COMPONENT>^"["GND{white}+clamp{white}+reference"]"	{ RET( COMPONENT_GND_CLAMP ) }

<S_COMPONENT>^"["Rload"]"					{ RET( COMPONENT_RLOAD ) }

<S_COMPONENT>^"["Derate{white}+VI"]"		{ RET( COMPONENT_DERATE_VI ) }

<S_COMPONENT>^"["Derate{white}+ramp"]"		{ RET( COMPONENT_DERATE_RAMP ) }

<S_COMPONENT>^"["R_pkg"]"					{ RET( COMPONENT_R_PKG ) }

<S_COMPONENT>^"["L_pkg"]"					{ RET( COMPONENT_L_PKG ) }

<S_COMPONENT>^"["C_pkg"]"					{ RET( COMPONENT_C_PKG ) }

<S_COMPONENT>^"["C_comp"]"					{ RET( COMPONENT_C_COMP ) }

<S_COMPONENT>^"["Sim{white}+time"]"			{ RET( COMPONENT_SIM_TIME ) }





<S_GET_VERSION>1\.1		{ BEGIN( INITIAL ); RET( VER_ONE_ONE ) }

<S_GET_VERSION>2\.1		{ BEGIN( INITIAL ); RET( VER_TWO_ONE ) }




<S_MODEL>^"["Model{white}+type"]"	{ 
	BEGIN( S_MODEL_TYPE );  
	RET( MODEL_TYPE ) 
	}

<S_MODEL>^"["C_comp"]"					{ RET( C_COMP ) }

<S_MODEL>^"["Polarity"]"				{ RET( POLARITY ) }

<S_MODEL>Inverting						{ RET( INVERTING ) }

<S_MODEL>Non-Inverting					{ RET( NON_INVERTING ) }

<S_MODEL>^"["Enable"]"					{ RET( ENABLE ) }

<S_MODEL>Active-High					{ RET( ACTIVE_HIGH ) }

<S_MODEL>Active-Low						{ RET( ACTIVE_LOW ) }

<S_MODEL>^"["Vinl"]"					{ RET( VINL ) }

<S_MODEL>^"["Vinh"]"					{ RET( VINH ) }

<S_MODEL>^"["Vmeas"]"					{ RET( VMEAS ) }

<S_MODEL>^"["Cref"]"					{ RET( CREF ) }

<S_MODEL>^"["Rref"]"					{ RET( RREF ) }

<S_MODEL>^"["Vref"]"					{ RET( VREF ) }

<S_MODEL>^"["Rgnd"]"					{ RET( RGND ) }

<S_MODEL>^"["Rpower"]"					{ RET( RPOWER ) }

<S_MODEL>^"["Rac"]"						{ RET( RAC ) }

<S_MODEL>^"["Cac"]"						{ RET( CAC ) }

<S_MODEL>^"["Temperature{white}+range"]"		{ RET( TEMP_RANGE ) }

<S_MODEL>^"["Voltage{white}+range"]"	{ RET( VOLTAGE_RANGE ) }

<S_MODEL>^"["Pullup{white}+reference"]"	{ RET( PULLUP ) }

<S_MODEL>^"["Pulldown{white}+reference"]"		{ RET( PULLDOWN ) }

<S_MODEL>^"["POWER{white}+Clamp{white}+reference"]"	{ RET( POWER_CLAMP ) }

<S_MODEL>^"["GND{white}+Clamp{white}+reference"]"	{ RET( GND_CLAMP ) }

<S_MODEL>^"["Rload"]"					{ RET( RLOAD ) }

<S_MODEL>^"["Rising{white}+waveform"]"	{ RET( RISING_WAVE ) }

<S_MODEL>^"["Falling{white}+waveform"]"	{ RET( FALLING_WAVE ) }

<S_MODEL>^"["Model{white}+file"]"		{ RET( MODEL_FILE ) }

<S_MODEL>^"["Derate{white}+VI"]"		{ RET( DERATE_VI ) }

<S_MODEL>^"["Derate{white}+ramp"]"		{ RET( DERATE_RAMP ) }

<S_MODEL>^"["Sim{white}+time"]"			{ RET( SIM_TIME ) }




<S_MODEL_TYPE>Input				{ BEGIN(S_MODEL); RET(INPUT) }

<S_MODEL_TYPE>Output			{ BEGIN(S_MODEL); RET(OUTPUT) }

<S_MODEL_TYPE>3"-"State			{ BEGIN(S_MODEL); RET(TRI_STATE) }

<S_MODEL_TYPE>Open_drain		{ BEGIN(S_MODEL); RET(OPEN_DRAIN) }

<S_MODEL_TYPE>I\/O_open_drain	{ BEGIN(S_MODEL); RET(I_O_OPEN_DRAIN) }

<S_MODEL_TYPE>I\/O_open_sink	{ BEGIN(S_MODEL); RET(I_O_OPEN_SINK) }

<S_MODEL_TYPE>I\/O_open_source	{ BEGIN(S_MODEL); RET(I_O_OPEN_SOURCE) }

<S_MODEL_TYPE>Input_ECL			{ BEGIN(S_MODEL); RET(INPUT_ECL) }

<S_MODEL_TYPE>I\/O_ECL			{ BEGIN(S_MODEL); RET(I_O_ECL) }

<S_MODEL_TYPE>I\/O				{ BEGIN(S_MODEL); RET(I_O) }

<S_MODEL_TYPE>Terminator		{ BEGIN(S_MODEL); RET(TERMINATOR) }

<S_MODEL_TYPE>Open_sink			{ BEGIN(S_MODEL); RET(OPEN_SINK) }

<S_MODEL_TYPE>Open_source		{ BEGIN(S_MODEL); RET(OPEN_SOURCE) }

<S_MODEL_TYPE>Output_ECL		{ BEGIN(S_MODEL); RET(OUTPUT_ECL) }


<S_PIN>^"->"								{ RET( CONT_LINE ) }


<S_SPICE_TYPE>HSpice {
	BEGIN( INITIAL );
	RET( HSPICE )
	}


<S_SPICE_TYPE>PSpice {
	BEGIN( INITIAL );
	RET( PSPICE )
	}

<S_SPICE_TYPE>Spice2 {
	BEGIN( INITIAL );
	RET( SPICE2 )
	}

<S_SPICE_TYPE>Spice3 {
	BEGIN( INITIAL );
	RET( SPICE3 )
	}

<S_SPICE_TYPE>Spectre {
	BEGIN( INITIAL );
	RET( SPECTRE )
	}


<*>POWER		{ RET(POWER) }
<*>NC			{ RET(NC) }
<*>NA			{ RET(NA) }
<*>GND			{ RET(GND) }

<*>=			{ RET('=') }


<INITIAL,S_PIN,S_MODEL,S_COMPONENT>{predigit}({float_core}|{digits})([a-zA-Z]+) {
	    	  char ccc;
		      int i = 0;
			  COPY_YYTEXT
	          lastNumberAsText = yylvalStr[idIndex];
			  
			  /* We have a scaled float.
			   * Get the first alphabetic character.  */
			   
			  while(lastNumberAsText[i++] < 'A');
			  switch(ccc = lastNumberAsText[i-1])
			    {
			    case 'a':
			      yylval.dval = 1.e-18;
			      break;
			    case 'f':
			      yylval.dval = 1.e-15;
			      break;
			    case 'p':
			      yylval.dval = 1.e-12;
			      break;
			    case 'n':
			      yylval.dval = 1.e-9;
			      break;
			    case 'u':
			      yylval.dval = 1.e-6;
			      break;
			    case 'm':
			      yylval.dval = 1.e-3;
			      break;
			    case 'k':
			      yylval.dval = 1.e3;
			      break;
			    case 'M':
			      yylval.dval = 1.e6;
			      break;
			    case 'G':
			      yylval.dval = 1.e9;
			      break;
			    case 'T':
			      yylval.dval = 1.e12;
			      break;

		            /* As in Spice ignore some units */

			    case 'V':
			    case 'A':
			      yylval.dval = 1.;
			      break;
			    }
			    lastNumberAsText[i] = 0;
			    yylval.dval *= atof(lastNumberAsText);
			    lastNumberAsText[i] = ccc;
			  RET(FLOATLIT) 
			  }

<INITIAL,S_PIN,S_MODEL,S_COMPONENT>{predigit}({float_core}{expon}?)|({digits}{expon})	{
			  COPY_YYTEXT
	                  lastNumberAsText = yylvalStr[idIndex];
			  yylval.dval = atof(yytext);
			  RET(FLOATLIT) 
			  }

<INITIAL,S_PIN,S_MODEL,S_COMPONENT>{predigit}{digits} {
			  yylval.ival = atoi(yytext);
			  COPY_YYTEXT
	          lastNumberAsText = yylvalStr[idIndex];
			  RET(INTLIT) 
			  }

<*>{not_eol_eof_white}+	{ 
			  COPY_YYTEXT
 			  yylval.id = yylvalStr[idIndex];
			  RET(IDENTIFIER) 
			  }

<*>{white}*

<*>{lf}"+"		{ ++lineNum; }

<*>({comment}([^\n]*))*{lf}	{ ++lineNum; RET(eol) }

<<EOF>>			{ return(eof); }


%%
