/* File *******************************************************************
*					North Carolina State University
*					Electronics Research Laboratory
*
* FILENAME :    s2ibslnk.h
*
* DESCRIPTION : This file implements the linked-list handling functions
*				used by s2ibis2.  They are implemented in the form of
*				macros which are inserted into the C code by the C
*				preprocessor.  This was done to make them maximally 
*				portable.
*
*
* FUNCTIONS:
*
*
* REVISION HISTORY :
*       Date        Programmer:         Description
*		8/24/95		Alan Glaser			Beta release to some members of the
*					Michael Steer		IBIS community
*
*
*H*/


/*
 * NEW_LINK
 *
 * Create a new link in a linked list of pointers and return the last
 * pointer in the list after alocating space for it.
 * Next_Pointer is modified to point to the last data structure in the list.
 */
#define \
   NEW_LINK(Top_Pointer_P, 	/* The address of the next pointer to be \
                                 * allocated. If this points to the first \
                                 * allocated (or even intermediate pointer) \
                                 * in the link list then the link is traversed \
                                 * to find the bottom of the linked list. */ \
            Next_Pointer_P,     /* Pointer_type *Next_Pointer; at the end this \
                                   will be the pointer to the last link in the \
                                   list unless there was an error in which case\
                                   it will be NULL */ \
            next_pointer_name, /* New_Pointer = Top_Pointer->Next_Pointer_name \
                                  Of course we must traverse the list down to \
                                  the bottom. */ \
            Pointer_type, \
            Type)  /* The pointer points to a data structure of type Type */ \
{ \
Pointer_type Old_Pointer; \
/* Find end of linked list */  \
for( *Next_Pointer_P = *Top_Pointer_P; *Next_Pointer_P != NULL; \
      Old_Pointer = *Next_Pointer_P, \
      *Next_Pointer_P = (*Next_Pointer_P)->next_pointer_name) ; \
/* Now allocate memory for the last link. */  \
if( ( *Next_Pointer_P = (Pointer_type) MALLOC(sizeof(Type)) ) == NULL) \
  yyerror("Out of memory"); \
if( *Top_Pointer_P == NULL) \
   *Top_Pointer_P = *Next_Pointer_P; \
else \
   Old_Pointer->next_pointer_name = *Next_Pointer_P; \
}

/*
 * DELETE_LINK
 *
 * Delete a link in a linked list of pointers and return
 *   Fix up pointers
 *   Set Link_Pointer
 */
#define \
   DELETE_LINK(Top_Pointer, \
            Link_pointer,       /* Pointer_type Link_pointer; This is the \
                                   pointer of the link to be removed. \
                                   On return Link_Pointer returns the \
                                   pointer to the link closest to the \
                                   poistion of the link deleted. \
                                   It will be null if there are no links. \
                                   will be the pointer to the last link in the \
                                   list unless there was an error in which case \
                                   it will be NULL */ \
            next_pointer_name, /* New_Pointer = Top_Pointer->Next_Pointer_name \
                                  We traverse the list this way. */ \
            Pointer_type) \
/* Traverse the linked list to find the link */ \
{ (Pointer_type) Next_Pointer, Old_Pointer; \
  for(Old_Pointer = Next_Pointer = (Top_Pointer); \
      (Next_Pointer == NULL) || (Next_Pointer == Link_Pointer); \
      Old_Pointer = Next_Pointer, \
        Next_Pointer = Next_Pointer->next_pointer_name); \
  if(Next_Pointer != NULL) \
        Old_pointer->(next_pointer_name) = Next_Pointer->(next_pointer_name); \
}


/*
 * DELETE_LINKED_LIST
 *
 * Delete a linked list
 *
 * The DELETE_ACTION must delete the contents of a link but not the 
 * the next link in the list.
 *
 */
#define \
   DELETE_LINKED_LIST(Top_Pointer_P, \
                               /* Pointer to top Pointer of linked list. \
                                  Everything from here down is deleted. \
                                  Top Pointer_P is NUll'ed */ \
            next_pointer_name, /* New_Pointer = Top_Pointer->Next_Pointer_name \
                                  We traverse the list this way. */ \
            DELETE_ACTION,     /* Action to be performed at each link */ \
            Pointer_type) \
{ \
Pointer_type Next_Pointer, Old_Pointer; \
/* We go through the links this way rather than recursively to be a ittle \
   bit faster. */ \
if(*Top_Pointer_P != NULL) \
  { \
  for(Old_Pointer = *Top_Pointer_P, \
        Next_Pointer = (*Top_Pointer_P)->next_pointer_name; \
      Next_Pointer != NULL ; \
      Next_Pointer = Next_Pointer->next_pointer_name) \
    { \
    DELETE_ACTION(Old_Pointer); \
    Old_Pointer->next_pointer_name = NULL; \
    FREE(Old_Pointer); \
    Old_Pointer = Next_Pointer; \
    } \
  FREE(*Top_Pointer_P); \
  } \
}



/*
 * TRAVERSE_LINKED_LIST
 *
 * Traverse a link in a linked list of pointers and 
 * perform action at each link.
 *
 */
#define \
   TRAVERSE_LINKED_LIST(Top_P, /* top pointer of linked list. */ \
            next_pointer_name, /* New_Pointer = Top_P->Next_Pointer_name \
                                  We traverse the list this way. */ \
            ACTION,     /* Action to be performed at each link */ \
            Pointer_type) \
{ \
Pointer_type Next_Pointer; \
if(Top_P != NULL) \
  { \
  for(Next_Pointer = Top_P; Next_Pointer != NULL ; \
      Next_Pointer = Next_Pointer->next_pointer_name) \
    ACTION(Next_Pointer); \
  } \
}

