/* File *******************************************************************
*					North Carolina State University
*					Electronics Research Laboratory
*
* FILENAME :    s2isutil.c
*
* DESCRIPTION : This file contains most of the utility routines for creating,
*				manipulating printing and destroying the various data
*				structures used by s2ibis2.
*
*
* FUNCTIONS:
*
*
* REVISION HISTORY :
*       Date        Programmer:         Description
*		8/24/95		Alan Glaser			Beta release to some members of the
*					Michael Steer		IBIS community
*
*
*H*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <sys/time.h>
#include <sys/types.h>
#include "s2ibis.h"
#include "s2ilink.h"

static int modelDefined, modelUsed, errorFlag;
static ibisModel_Pt models_P;
static ibisPin_Pt   pins_P;
static char *globalName;

void create_String( char *, char ** );

int create_Model( ibisModel_Pt *, ibisModel_Pt * );

int create_Pin( ibisPin_Pt *, ibisPin_Pt * );

int create_VITable( ibisVItable_Pt *, int );

int create_Component( ibisComponent_Pt *, ibisComponent_Pt * );

int create_TOP( ibisTOP_Pt *, char * );

int ibis_Delete( ibisTOP_Pt * );

#define PRINT_STRING_DEFAULT	0
#define PRINT_INDENT_ONCE		1


/*
 * Print utility routines
 */

void
spaceFill(	INOUT	char	*string,
			IN		int		numSpaces )

{
int i;

for ( i = 0; i < numSpaces; i++ )
	strcat( string, " " );
}


void nullString(	INOUT	char	*string,
					IN		int		size )

{
int i;
for ( i = 0; i < size; i++ )
	string[i] = '\0';
}



/* Function ***************************************************************
*
* NAME : fprintString
*
* INPUTS :	FILE		*outFile	-- ptr to file to print to
*			char		*indent		-- string to print at beginning of line
*			int			width		-- with of string (indent + string)
*			char		*string		-- string to print in an indented
*									   ragged-right format
*			int			option		-- one of:
*									   PRINT_STRING_DEFAULT - print indent
*									   		on every line
*									   PRINT_INDENT_ONCE - print indent on
*									   		first line only, but indent lines
*											2 - end with equivalent whitespace;
*											terminate last line with line feed.
*			ibisPin_Pt	currentPin_P	ptr to current component
*			ibisPin_Pt	pinList_P		ptr to global pin list
*			int			hasPinMapping	flag for existence of pin map
*
*F*/

void 
fprintString(	IN 	FILE	*outFile,
				IN	char 	*indent, 
				IN 	int 	width, 
				IN 	char 	*string, 
				IN 	int 	option)

{
int 	start = 0,
		i = 0,
		j = 0,
		indentLength = 0,
		indent_printed = 0,
		length,	/* length of string */
		width_less_indent;

char	hold;	/* holding character */


if ( indent != NULL )
	for ( j = 0; j < strlen( indent ); j ++ )
		if ( indent[j] == '\t' )
			indentLength += 8 - ( indentLength % 8 );
		else
			indentLength++;

width_less_indent = width - indentLength;
width_less_indent -= 1;  /* account for blank between indent and string */

if ( string != NULL )
	length = strlen( string );
else
	length = 0;

while ( ( start+i < length ) || !( indent_printed ) )
	{
	while ( i < width_less_indent && string[start+i] ) 
		i++;

    if ( indent_printed & ( option == PRINT_INDENT_ONCE ) )
    	{
    	for( j = 0; j < indentLength; j++ )
        	fprintf( outFile, " " );
    	}
	else
		{
		if ( indent != NULL )
			if ( indent[0] != 0 )
				fprintf( outFile, "%s", indent );
		indent_printed = 1;
		}

	if ( i < width_less_indent )
		{
		if ( string != NULL )
			{
			if (string[start] != 0 )
				fprintf( outFile, "%s\n", &string[start] );
			else
				fprintf( outFile, "\n" );
			}
		else
			fprintf( outFile, "\n" );
		start = length;
		break;
		}
	else
		{
		/* find the last blank */
		while ( i && ( string[start+i] != ' ' ) )
			i--;
		hold = string[start+i];
		string[start+i] = 0;
		if ( indentLength )
			fprintf( outFile, "%s\n", &string[start] );
		else
			fprintf( outFile, "%s\n", &string[start] );
		string[start+i] = hold;
		start += i + 1;
		i = 0;
		}
	}
}


/* Function ***************************************************************
*
* NAME : isExponent
*
* INPUTS : 	double	number	--	number to be tested
*			double	expon	--	possible exponent
*
*
* OUTPUTS:	0 if expon is not exponent
*			1 if expon is exponent
*
*
* NOTES :	divides number by expon, and checks to see if 1<=number<10
*			Note: This is supposed to work with engineering units, so
*			you should only feed it exponents spaced by 1e3
*
*F*/

int
isExponent(	IN	double	number,
			IN	double	expon )

{
if ( ( fabs( number / expon ) >= 1 ) && ( fabs( number / expon ) < 1000 ) )
	return( 1 );
else
	return( 0 );
}


/*
 *
 * Get_Double_as_String
 *
 * Print a number as g or NA if number == USE_NA
 */
void Get_Double_as_String(double number, char *s, char *trailStr)
{
if ( IS_USE_NA( number ) )
	sprintf( s, "  NA     " );
else
	{
	if ( isExponent( number, 1e-15 ) )
		sprintf( s, "%7.2f%s%s", number/1e-15, "f", trailStr );
	else if ( isExponent( number, 1e-12 ) )
		sprintf( s, "%7.2f%s%s", number/1e-12, "p", trailStr );
	else if ( isExponent( number, 1e-9 ) )
		sprintf( s, "%7.2f%s%s", number/1e-9, "n", trailStr );
	else if ( isExponent( number, 1e-6 ) )
		sprintf( s, "%7.2f%s%s", number/1e-6, "u", trailStr );
	else if ( isExponent( number, 1e-3 ) )
		sprintf( s, "%7.2f%s%s", number/1e-3, "m", trailStr );
	else if ( isExponent( number, 1e3 ) )
		sprintf( s, "%7.2f%s%s", number/1e3, "k", trailStr );
	else if ( isExponent( number, 1e6 ) )
		sprintf( s, "%7.2f%s%s", number/1e6, "M", trailStr );
	else if ( isExponent( number, 1e9 ) )
		sprintf( s, "%7.2f%s%s", number/1e9, "G", trailStr );
	else if ( isExponent( number, 1e12 ) )
		sprintf( s, "%7.2f%s%s", number/1e12, "T", trailStr );
	else
		sprintf( s, "%7.2f%s", number, trailStr );
	}
}

/*
 *
 * Get_Double_as_LString
 *
 * Print a number as g or NA if number == USE_NA (left justified)
 */
void Get_Double_as_LString(double number, char *s, char *trailStr)
{
char 	dummyBuffer[128];
int 	i;

if ( IS_USE_NA( number ) )
	sprintf( s, "NA" );
else
	{
	if ( isExponent( number, 1e-15 ) )
		{
		sprintf( s, "%-7.2f", number/1e-15 );
		for ( i = 0; i < 7 ; i++ )
			if ( s[i] == ' ' )
				s[i] = '\0';
		sprintf( dummyBuffer, "%s%s", "f", trailStr );	
		strcat( s, dummyBuffer );
		}
	else if ( isExponent( number, 1e-12 ) )
		{
		sprintf( s, "%-7.2f", number/1e-12 );
		for ( i = 0; i < 7 ; i++ )
			if ( s[i] == ' ' )
				s[i] = '\0';
		sprintf( dummyBuffer, "%s%s", "p", trailStr );	
		strcat( s, dummyBuffer );
		}
	else if ( isExponent( number, 1e-9 ) )
		{
		sprintf( s, "%-7.2f", number/1e-9 );
		for ( i = 0; i < 7 ; i++ )
			if ( s[i] == ' ' )
				s[i] = '\0';
		sprintf( dummyBuffer, "%s%s", "n", trailStr );	
		strcat( s, dummyBuffer );
		}
	else if ( isExponent( number, 1e-6 ) )
		{
		sprintf( s, "%-7.2f", number/1e-6 );
		for ( i = 0; i < 7 ; i++ )
			if ( s[i] == ' ' )
				s[i] = '\0';
		sprintf( dummyBuffer, "%s%s", "u", trailStr );	
		strcat( s, dummyBuffer );
		}
	else if ( isExponent( number, 1e-3 ) )
		{
		sprintf( s, "%-7.2f", number/1e-3 );
		for ( i = 0; i < 7 ; i++ )
			if ( s[i] == ' ' )
				s[i] = '\0';
		sprintf( dummyBuffer, "%s%s", "m", trailStr );	
		strcat( s, dummyBuffer );
		}
	else if ( isExponent( number, 1e3 ) )
		{
		sprintf( s, "%-7.2f", number/1e3 );
		for ( i = 0; i < 7 ; i++ )
			if ( s[i] == ' ' )
				s[i] = '\0';
		sprintf( dummyBuffer, "%s%s", "k", trailStr );	
		strcat( s, dummyBuffer );
		}
	else if ( isExponent( number, 1e6 ) )
		{
		sprintf( s, "%-7.2f", number/1e6 );
		for ( i = 0; i < 7 ; i++ )
			if ( s[i] == ' ' )
				s[i] = '\0';
		sprintf( dummyBuffer, "%s%s", "M", trailStr );	
		strcat( s, dummyBuffer );
		}
	else if ( isExponent( number, 1e9 ) )
		{
		sprintf( s, "%-7.2f", number/1e9 );
		for ( i = 0; i < 7 ; i++ )
			if ( s[i] == ' ' )
				s[i] = '\0';
		sprintf( dummyBuffer, "%s%s", "G", trailStr );	
		strcat( s, dummyBuffer );
		}
	else if ( isExponent( number, 1e12 ) )
		{
		sprintf( s, "%-7.2f", number/1e12 );
		for ( i = 0; i < 7 ; i++ )
			if ( s[i] == ' ' )
				s[i] = '\0';
		sprintf( dummyBuffer, "%s%s", "T", trailStr );	
		strcat( s, dummyBuffer );
		}
	else
		{
		sprintf( s, "%-7.2f", number );
		for ( i = 0; i < 7 ; i++ )
			if ( s[i] == ' ' )
				s[i] = '\0';
		sprintf( dummyBuffer, "%s", trailStr );	
		strcat( s, dummyBuffer );
		}
	}
}


/*
 *
 * Get_TypMinMax_header_as_String(s);
 *
 */
Get_TypMinMax_header_as_String(char *s)
{
static char header[]="  I(typ)              I(min)              I(max)";
int i;

for(i = 0; (s[i] = header[i]) != 0; i++);
}

/*
 *
 * Get_VTypMinMax_header_as_String(s);
 *
 */
Get_VTypMinMax_header_as_String(char *s)
{
static char Vheader[]="  V(typ)              V(min)              V(max)";
int i;

for(i = 0; (s[i] = Vheader[i]) != 0; i++);
}


void
printGenericTypMinMaxHeader( IN FILE *outFile )
{
fprintString( outFile, 
			  "| variable       typ                 min                 max", 
			  IBIS_PRINT_WIDTH, "", PRINT_INDENT_ONCE );
}


/*
 *
 * Get_TypMinMax_as_String
 *
 * Convert the typ min max sequence to a string
 */
Get_TypMinMax_as_String(	IN 		ibisTypMinMax_t TypMinMax, 
							INOUT 	char 			*s,
                        	IN 		int 			max_string_length,
							IN		char			*trailStr )
{
int position;
char dummyString[16];

if(max_string_length > 45)
  {
  Get_Double_as_String(TypMinMax.typ, s, trailStr);

  position = strlen(s);
  spaceFill( s, 20 - position );
  Get_Double_as_String(TypMinMax.min, dummyString, trailStr);
  strcat( s, dummyString );

  position = strlen(s);
  spaceFill( s, 40 - position );
  Get_Double_as_String(TypMinMax.max, dummyString, trailStr);
  strcat( s, dummyString );

  }
else
  {
  int i;
  for(i = max_string_length; i != 0; i--)
     s[i-1] = '?';
  return(1);
  }
return( 0 );
}


/*
 *
 * Get_TypMinMax_as_LString
 *
 * Convert the typ min max sequence to a left-justified string
 */
Get_TypMinMax_as_LString(	IN 		ibisTypMinMax_t TypMinMax, 
							INOUT 	char 			*s,
                        	IN 		int 			max_string_length,
							IN		char			*trailStr )
{
int position;
char dummyString[16];

if(max_string_length > 45)
  {
  Get_Double_as_LString(TypMinMax.typ, s, trailStr);

  position = strlen(s);
  spaceFill( s, 20 - position );
  Get_Double_as_LString(TypMinMax.min, dummyString, trailStr);
  strcat( s, dummyString );

  position = strlen(s);
  spaceFill( s, 40 - position );
  Get_Double_as_LString(TypMinMax.max, dummyString, trailStr);
  strcat( s, dummyString );

  }
else
  {
  int i;
  for(i = max_string_length; i != 0; i--)
     s[i-1] = '?';
  return(1);
  }
return( 0 );
}



/* Diff Pin Objects */

/* Function ***************************************************************
*
* NAME : createDiffPin
*
* INPUTS :  ibisDiffPin_Pt	*top_Ponter_PP -- ptr to existing diff pin
*
*
* OUTPUTS : 0 if successful
*			1 if not
*
*
* RETURNS :  *ibisDiffPin_Pt -- handle to new diff pin entry object
*
*F*/

int
createDiffPin(	IN  ibisDiffPin_Pt *top_Pointer_PP,
           		OUT ibisDiffPin_Pt *diffPin_PP ) 

{
NEW_LINK( top_Pointer_PP, diffPin_PP, nextDiffPin_P, ibisDiffPin_Pt, 
		  ibisDiffPin_t)
return( *diffPin_PP == NULL);
}


/* Function ***************************************************************
*
* NAME : deleteDiffPin
*
* INPUTS :  ibisDiffPin_Pt	DiffPin_P -- ptr to diff pin to be deleted
*
*
* NOTES :	Doesn't really do anything, but necessary for DELETE_LINKED_LIST
*
*F*/

void
deleteDiffPin( IN  ibisDiffPin_Pt diffPin_P ) 

{
return;
}

/* Function ***************************************************************
*
* NAME : printDiffPin
* 
* INPUTS : ibisDiffPin_Pt currentDiffPin_P
*	
* 
* NOTES / SIDE EFFECTS : Prints the diff pin entry
*
*
*F*/

void
printDiffPin( IN	FILE			*outFile,
			  IN	ibisDiffPin_Pt	currentDiffPin_P )
{

/*==== Declarations =====================================================*/ 
char	tempBuffer[80];	/* holds the output string as it's built */
char	numBuffer[MAX_R_PIN_STRING_LENGTH + 1];


/*==== Code =============================================================*/
sprintf( tempBuffer, "%s", currentDiffPin_P->pinName );
spaceFill( tempBuffer, 12 - strlen( tempBuffer ) );
strcat( tempBuffer, currentDiffPin_P->invPinName );
spaceFill( tempBuffer, 21 - strlen( tempBuffer ) );
Get_Double_as_String( currentDiffPin_P->vdiff, numBuffer, "V" );
strcat( tempBuffer, numBuffer );
spaceFill( tempBuffer, 31 - strlen( tempBuffer ) );
Get_Double_as_String( currentDiffPin_P->tdelayTyp, numBuffer, "s" );
strcat( tempBuffer, numBuffer );
if ( !( IS_NOT_USED( currentDiffPin_P->tdelayMin ) ) )
	{
	spaceFill( tempBuffer, 43 - strlen( tempBuffer ) );
	Get_Double_as_String( currentDiffPin_P->tdelayMin, numBuffer, "s" );
	strcat( tempBuffer, numBuffer );
	spaceFill( tempBuffer, 55 - strlen( tempBuffer ) );
	Get_Double_as_String( currentDiffPin_P->tdelayMax, numBuffer, "s" );
	strcat( tempBuffer, numBuffer );
	}

fprintString( outFile, "", IBIS_PRINT_WIDTH, tempBuffer, PRINT_INDENT_ONCE );

}




/* Pin Map Objects */

/* Function ***************************************************************
*
* NAME : createPinMap
*
* INPUTS :  ibisPinMap_Pt	*top_Ponter_PP -- ptr to existing pin map
*
*
* OUTPUTS : 0 if successful
*			1 if not
*
*
* RETURNS :  *ibisPinMap_Pt -- handle to new pin map entry object
*
*F*/

int
createPinMap(	IN  ibisPinMap_Pt *top_Pointer_PP,
           		OUT ibisPinMap_Pt *pinMap_PP ) 

{
NEW_LINK( top_Pointer_PP, pinMap_PP, nextPinMap_P, ibisPinMap_Pt, ibisPinMap_t)
return( *pinMap_PP == NULL);
}


/* Function ***************************************************************
*
* NAME : getMatchingPinMap
*
* INPUTS :  char   			*searchName  -- pointer to pin name to find
*			ibisPinMap_Pt	pinMapList_P -- pointer to pin list to search
*
*
* OUTPUTS :  ibisPinMap_Pt -- points to pin with name searchName
*
*
* NOTES / SIDE EFFECTS : Finds pin map with name searchName.  Returns null if
*						 it does not exist.
*
*F*/

ibisPinMap_Pt
getMatchingPinMap(	IN	char			*searchName,
					IN	ibisPinMap_Pt	pinMapList_P )

{
ibisPinMap_Pt	tempPinMap_P = pinMapList_P;

/*------------------------------------------------------------------------
*
*	If the search name is the null string just return the null pointer 
*
*-----------------------------------------------------------------------*/
if ( !( strcmp( searchName, "" ) ) )
	return( NULL );

/*------------------------------------------------------------------------
*
*	Otherwise, search through the pin map list until you find the matching
*	pin map.
*
*-----------------------------------------------------------------------*/
while( ( tempPinMap_P != NULL )  && 
	   strcmp( searchName, tempPinMap_P->pinName ) )
	tempPinMap_P = tempPinMap_P->nextPinMap_P;

/*------------------------------------------------------------------------
*
*	If you reach the end of the list without finding a match, report it.
*
*-----------------------------------------------------------------------*/
if ( tempPinMap_P == NULL )
	{
	fprintf( stderr, "s2ibis2: pin %s not in pin map.\n",
			 searchName );
	fprintf( stderr, "         Please include the pin in the " );
	fprintf( stderr, "[Pin Mapping] section of your\n" );
	fprintf( stderr, "         input file, or make sure the pin " );
	fprintf( stderr, "name is not misspelled.\n" );
	}

return( tempPinMap_P );
}


/* Function ***************************************************************
*
* NAME : deletePinMap
*
* INPUTS :  ibisPinMap_Pt	pinMap_P -- ptr to pin map to be deleted
*
*
* NOTES :	Doesn't really do anything, but necessary for DELETE_LINKED_LIST
*
*F*/

void
deletePinMap( IN  ibisPinMap_Pt pinMap_P ) 

{
return;
}

/* Function ***************************************************************
*
* NAME : printPinMap
* 
* INPUTS : ibisPin_Pt currentPinMap_P
*	
* 
* NOTES / SIDE EFFECTS : Prints the pin mapping for a single pin
*
*
*F*/

void
printPinMap(	IN	FILE			*outFile,
				IN	ibisPinMap_Pt	currentPinMap_P )
{

/*==== Declarations =====================================================*/ 
char	tempBuffer[80];		/* holds the output string as it's constructed */


/*==== Code =============================================================*/
sprintf( tempBuffer, "%s", currentPinMap_P->pinName );
spaceFill( tempBuffer, 15 - strlen( tempBuffer ) );
strcat( tempBuffer, currentPinMap_P->pulldownRef );
spaceFill( tempBuffer, 31 - strlen( tempBuffer ) );
strcat( tempBuffer, currentPinMap_P->pullupRef );
spaceFill( tempBuffer, 47 - strlen( tempBuffer ) );
strcat( tempBuffer, currentPinMap_P->gndClampRef );
spaceFill( tempBuffer, 63 - strlen( tempBuffer ) );
strcat( tempBuffer, currentPinMap_P->powerClampRef );

fprintString( outFile, "", IBIS_PRINT_WIDTH, tempBuffer, PRINT_INDENT_ONCE );
}



/* Model Objects */

/* Function ***************************************************************
*
* NAME : getMatchingModel
*
* INPUTS :  char   			*searchName -- pointer to model name to find
*			ibisModel_Pt	modelList_P -- pointer to model list to search
*
*
* OUTPUTS :  ibisModel_Pt -- points to model with name searchName
*
*
* NOTES / SIDE EFFECTS : Finds model with name searchName.  Returns null if
*						 it does not exist.
*
*F*/

ibisModel_Pt
getMatchingModel(	IN	char			*searchName,
					IN	ibisModel_Pt	modelList_P )

{
ibisModel_Pt	tempModel_P = modelList_P;

/*------------------------------------------------------------------------
*
*	If the search name is the null string just return the null pointer 
*
*-----------------------------------------------------------------------*/
if ( !( strcmp( searchName, "" ) ) )
	return( NULL );

/*------------------------------------------------------------------------
*
*	Otherwise, search through the model list until you find the matching
*	model.
*
*-----------------------------------------------------------------------*/
while( ( tempModel_P != NULL )  && 
		 strcmp( searchName, tempModel_P->modelName ) )
	tempModel_P = tempModel_P->nextModel_P;

/*------------------------------------------------------------------------
*
*	If you reach the end of the list without finding a match, report it.
*
*-----------------------------------------------------------------------*/
if ( tempModel_P == NULL )
	{
	fprintf( stderr, "s2ibis2: Model %s not found in model list.\n",
			 searchName );
	}

return( tempModel_P );
}



/*
 * create_Model
 *
 * Create model data structure and NULL all entries.
 */
create_Model(IN  ibisModel_Pt *top_Pointer_PP,/* This must already exist
                                               * and *top_Pointer_PP
                                               * must point to the first
                                               * model in the linked list.
                                               */
             OUT ibisModel_Pt *model_PP       /* This is the pointer
                                               * to the new entry. */
             )
{
ibisModel_Pt	model_P;

NEW_LINK( top_Pointer_PP, model_PP, nextModel_P, ibisModel_Pt, ibisModel_t);
if ( *model_PP != NULL )
	{
	model_P = *model_PP;
	model_P->Vinl = USE_NA;
	model_P->Vinh = USE_NA;
	model_P->Vmeas = USE_NA;
	model_P->Cref = USE_NA;
	model_P->Rref = USE_NA;
	model_P->Vref = USE_NA;
	model_P->Rload = USE_NA;
	model_P->derateVIPct = USE_NA;
	model_P->derateRampPct = USE_NA;
	model_P->simTime = USE_NA;
	model_P->c_comp.typ = USE_NA;
	model_P->c_comp.min = USE_NA;
	model_P->c_comp.max = USE_NA;
	model_P->tempRange.typ = USE_NA;
	model_P->tempRange.min = USE_NA;
	model_P->tempRange.max = USE_NA;
	model_P->voltageRange.typ = USE_NA;
	model_P->voltageRange.min = USE_NA;
	model_P->voltageRange.max = USE_NA;
	model_P->pullupRef.typ = USE_NA;
	model_P->pullupRef.min = USE_NA;
	model_P->pullupRef.max = USE_NA;
	model_P->pulldownRef.typ = USE_NA;
	model_P->pulldownRef.min = USE_NA;
	model_P->pulldownRef.max = USE_NA;
	model_P->powerClampRef.typ = USE_NA;
	model_P->powerClampRef.min = USE_NA;
	model_P->powerClampRef.max = USE_NA;
	model_P->gndClampRef.typ = USE_NA;
	model_P->gndClampRef.min = USE_NA;
	model_P->gndClampRef.max = USE_NA;
	model_P->Rgnd.typ = USE_NA;
	model_P->Rgnd.min = USE_NA;
	model_P->Rgnd.max = USE_NA;
	model_P->Rpower.typ = USE_NA;
	model_P->Rpower.min = USE_NA;
	model_P->Rpower.max = USE_NA;
	model_P->Rac.typ = USE_NA;
	model_P->Rac.min = USE_NA;
	model_P->Rac.max = USE_NA;
	model_P->Cac.typ = USE_NA;
	model_P->Cac.min = USE_NA;
	model_P->Cac.max = USE_NA;
	model_P->ramp.dv_r.typ = USE_NA;
	model_P->ramp.dv_r.min = USE_NA;
	model_P->ramp.dv_r.max = USE_NA;
	model_P->ramp.dt_r.typ = USE_NA;
	model_P->ramp.dt_r.min = USE_NA;
	model_P->ramp.dt_r.max = USE_NA;
	model_P->ramp.dv_f.typ = USE_NA;
	model_P->ramp.dv_f.min = USE_NA;
	model_P->ramp.dv_f.max = USE_NA;
	model_P->ramp.dt_f.typ = USE_NA;
	model_P->ramp.dt_f.min = USE_NA;
	model_P->ramp.dt_f.max = USE_NA;
	}

return( *model_PP == NULL);
}


/* Function ***************************************************************
*
* NAME : copyComponentDataToModels
*
* INPUTS :	ibisTOP_Pt		top_P		ptr to top IBIS structure
*
*
* RETURNS : ibisTOP_Pt		top_P		ptr to top IBIS structure
*
*
* NOTES / SIDE EFFECTS :	This routine copies the global data found in the
*							component to its models, if it is not already 
*							there.
*F*/

void
copyComponentDataToModels( INOUT	ibisTOP_Pt		top_P )

{
ibisComponent_Pt	currentComponent_P;
ibisModel_Pt		currentModel_P;

for ( currentComponent_P = top_P->components_P;
	  currentComponent_P != NULL;
	  currentComponent_P = currentComponent_P->nextComponent_P )
	for ( currentModel_P = currentComponent_P->models_P;
		  currentModel_P != NULL;
		  currentModel_P = currentModel_P->nextModel_P )
		{
		if ( IS_USE_NA( currentModel_P->c_comp.typ ) )
			currentModel_P->c_comp = currentComponent_P->c_comp;
		if ( IS_USE_NA( currentModel_P->tempRange.typ ) )
			currentModel_P->tempRange = currentComponent_P->tempRange;
		if ( IS_USE_NA( currentModel_P->voltageRange.typ ) &&
			 IS_USE_NA( currentModel_P->pullupRef.typ ) )
			currentModel_P->voltageRange = currentComponent_P->voltageRange;
		if ( IS_USE_NA( currentModel_P->pullupRef.typ ) &&
			 IS_USE_NA( currentModel_P->voltageRange.typ ) )
			currentModel_P->pullupRef = currentComponent_P->pullupRef;
		if ( IS_USE_NA( currentModel_P->pulldownRef.typ ) &&
			 IS_USE_NA( currentModel_P->voltageRange.typ ) )
			currentModel_P->pulldownRef = currentComponent_P->pulldownRef;
		if ( IS_USE_NA( currentModel_P->powerClampRef.typ ) &&
			 IS_USE_NA( currentModel_P->voltageRange.typ ) )
			currentModel_P->powerClampRef = currentComponent_P->powerClampRef;
		if ( IS_USE_NA( currentModel_P->gndClampRef.typ ) &&
			 IS_USE_NA( currentModel_P->voltageRange.typ ) )
			currentModel_P->gndClampRef = currentComponent_P->gndClampRef;
		if ( IS_USE_NA( currentModel_P->Rload ) )
			currentModel_P->Rload = currentComponent_P->Rload;
		if ( IS_USE_NA( currentModel_P->derateVIPct ) )
			currentModel_P->derateVIPct = currentComponent_P->derateVIPct;
		if ( IS_USE_NA( currentModel_P->derateRampPct ) )
			currentModel_P->derateRampPct = currentComponent_P->derateRampPct;
		if ( IS_USE_NA( currentModel_P->simTime ) )
			currentModel_P->simTime = currentComponent_P->simTime;
		}
}






/* Function ***************************************************************
*
* NAME : printModelHead
*
* INPUTS :	char	*modelName		Name of model
*
*
* OUTPUTS : prints header lines for the model    
*
*F*/

void
printModelHead(	IN	FILE	*outFile,
				IN	char	*modelName	)

{
char	tempBuffer[IBIS_PRINT_WIDTH];
int		spacing;

nullString( tempBuffer, IBIS_PRINT_WIDTH );
fprintString( outFile, "|", IBIS_PRINT_WIDTH, 
	"************************************************************************",
	PRINT_INDENT_ONCE );
spacing = ( 72 - ( strlen( modelName ) + strlen( "Model " ) ) ) / 2 ;
spaceFill( tempBuffer, spacing );
strcat( tempBuffer, "Model " );
strcat( tempBuffer, modelName );
fprintString( outFile, "|", IBIS_PRINT_WIDTH, tempBuffer, PRINT_INDENT_ONCE );
fprintString( outFile, "|", IBIS_PRINT_WIDTH, 
	"************************************************************************",
	PRINT_INDENT_ONCE );
fprintString( outFile, "|", IBIS_PRINT_WIDTH, "", PRINT_INDENT_ONCE );
}



/*
 * print_Model
 *
 * Print model data structure.
 */
print_Model( IN	FILE			*outFile,
			 IN	ibisModel_Pt	model_P )
{
ibisWaveTable_Pt	risingWave_P, 
					fallingWave_P;
char s[MAX_IBIS_STRING_LENGTH], *s_P;
char numberBuffer[128];

numberBuffer[0] = 0;
nullString( s, MAX_IBIS_STRING_LENGTH );
printModelHead( outFile, model_P->modelName );
fprintString( outFile, "[Model]          ", IBIS_PRINT_WIDTH, 
			  model_P->modelName, PRINT_INDENT_ONCE );

/* model type */
switch(model_P->modelType)
  {
  case MODEL_TYPE_INPUT:
    s_P = "Input";
    break;
  case MODEL_TYPE_OUTPUT:
    s_P = "Output";
    break;
  case MODEL_TYPE_3_STATE:
    s_P = "3-state";
    break;
  case MODEL_TYPE_OPEN_DRAIN:
    s_P = "Open_Drain";
    break;
  case MODEL_TYPE_I_O:
    s_P = "I/O";
    break;
  case MODEL_TYPE_I_O_OPEN_DRAIN:
    s_P = "I/O_open_drain";
    break;
  case MODEL_TYPE_I_O_OPEN_SINK:
    s_P = "I/O_open_sink";
    break;
  case MODEL_TYPE_I_O_OPEN_SOURCE:
    s_P = "I/O_open_source";
    break;
  case MODEL_TYPE_INPUT_ECL:
    s_P = "Input_ECL";
    break;
  case MODEL_TYPE_I_O_ECL:
    s_P = "I/O_ECL";
    break;
  case MODEL_TYPE_TERMINATOR:
    s_P = "Terminator";
    break;
  case MODEL_TYPE_OPEN_SINK:
    s_P = "Open_sink";
    break;
  case MODEL_TYPE_OPEN_SOURCE:
    s_P = "Open_source";
    break;
  case MODEL_TYPE_OUTPUT_ECL:
    s_P = "Output_ECL";
    break;
  case MODEL_TYPE_UNDEFINED:
  default:
    s_P = "?? undefined ??";
    break;
  }
fprintString( outFile, "Model_type       ", IBIS_PRINT_WIDTH, s_P, 
			  PRINT_STRING_DEFAULT );



/* polarity */
switch(model_P->polarity)
  {
  case MODEL_POLARITY_NON_INVERTING:
    s_P = "Non-Inverting";
    break;
  case MODEL_POLARITY_INVERTING:
    s_P = "Inverting";
    break;
  case MODEL_TYPE_UNDEFINED:
  default:
    s_P = "";
    break;
  }
if ( strcmp( s_P, "" ) )
	fprintString( outFile, "Polarity         ", IBIS_PRINT_WIDTH, s_P, 
				  PRINT_STRING_DEFAULT );



/* enable */
switch(model_P->enable)
  {
  case MODEL_ENABLE_ACTIVE_HIGH:
    s_P = "Active-High";
    break;
  case MODEL_ENABLE_ACTIVE_LOW:
    s_P = "Active-Low";
    break;
  case MODEL_TYPE_UNDEFINED:
  default:
    s_P = "";
    break;
  }
if ( strcmp( s_P, "" ) )
	fprintString( outFile, "Enable           ", IBIS_PRINT_WIDTH, s_P, 
				  PRINT_STRING_DEFAULT);


if ( !( IS_USE_NA( model_P->Vinl ) ) )
	{
	Get_Double_as_String( model_P->Vinl, s, "V" );
	fprintString( outFile, "Vinl =", IBIS_PRINT_WIDTH, s, 
				  PRINT_STRING_DEFAULT);
	}

if ( !( IS_USE_NA( model_P->Vinh ) ) )
	{
	Get_Double_as_String( model_P->Vinh, s, "V" );
	fprintString( outFile, "Vinh =", IBIS_PRINT_WIDTH, s, 
				  PRINT_STRING_DEFAULT);
	}

if ( !( IS_USE_NA( model_P->Vmeas ) ) )
	{
	Get_Double_as_String( model_P->Vmeas, s, "V" );
	fprintString( outFile, "Vmeas =", IBIS_PRINT_WIDTH, s, 
				  PRINT_STRING_DEFAULT);
	}

if ( !( IS_USE_NA( model_P->Cref ) ) )
	{
	Get_Double_as_String( model_P->Cref, s, "F" );
	fprintString( outFile, "Cref =", IBIS_PRINT_WIDTH, s, 
				  PRINT_STRING_DEFAULT);
	}

if ( !( IS_USE_NA( model_P->Rref ) ) )
	{
	Get_Double_as_String( model_P->Rref, s, "" );
	fprintString( outFile, "Rref =", IBIS_PRINT_WIDTH, s, 
				  PRINT_STRING_DEFAULT);
	}

if ( !( IS_USE_NA( model_P->Vref ) ) )
	{
	Get_Double_as_String( model_P->Vref, s, "V" );
	fprintString( outFile, "Vref =", IBIS_PRINT_WIDTH, s, 
				  PRINT_STRING_DEFAULT);
	}

Get_TypMinMax_as_LString(model_P->c_comp, s, IBIS_PRINT_WIDTH-8, "F" );
fprintString( outFile, "C_comp           ", IBIS_PRINT_WIDTH, s, 
			  PRINT_STRING_DEFAULT);

if ( !( IS_USE_NA( model_P->Rgnd.typ ) ) )
	{
	Get_TypMinMax_as_LString(model_P->Rgnd, s, IBIS_PRINT_WIDTH-8, "");
  	fprintString( outFile, "[Rgnd]           ", IBIS_PRINT_WIDTH, s, 
				  PRINT_STRING_DEFAULT );
	}
	
if ( !( IS_USE_NA( model_P->Rpower.typ ) ) )
	{
	Get_TypMinMax_as_LString(model_P->Rpower, s, IBIS_PRINT_WIDTH-8, "");
  	fprintString( outFile, "[Rpower]         ", IBIS_PRINT_WIDTH, s, 
				  PRINT_STRING_DEFAULT );
	}
	
if ( !( IS_USE_NA( model_P->Rac.typ ) ) )
	{
	Get_TypMinMax_as_LString(model_P->Rac, s, IBIS_PRINT_WIDTH-8, "");
  	fprintString( outFile, "[Rac]            ", IBIS_PRINT_WIDTH, s, 
				  PRINT_STRING_DEFAULT );
	}
	
if ( !( IS_USE_NA( model_P->Cac.typ ) ) )
	{
	Get_TypMinMax_as_LString(model_P->Cac, s, IBIS_PRINT_WIDTH-8, "F");
  	fprintString( outFile, "[Cac]            ", IBIS_PRINT_WIDTH, s, 
				  PRINT_STRING_DEFAULT );
	}
	
fprintString( outFile, "|", IBIS_PRINT_WIDTH, "", PRINT_STRING_DEFAULT );
fprintString( outFile, "|", IBIS_PRINT_WIDTH, "", PRINT_STRING_DEFAULT );

if ( !( IS_USE_NA( model_P->tempRange.typ ) ) )
	{
	Get_TypMinMax_as_LString(model_P->tempRange, s, IBIS_PRINT_WIDTH-8, "");
  	fprintString( outFile, "[Temperature Range]       ", IBIS_PRINT_WIDTH, s, 
				  PRINT_STRING_DEFAULT );
	}
	
if ( !( IS_USE_NA( model_P->voltageRange.typ ) ) )
	{
	Get_TypMinMax_as_LString(model_P->voltageRange, s, IBIS_PRINT_WIDTH-8, "V");
  	fprintString( outFile, "[Voltage Range]           ", IBIS_PRINT_WIDTH, s, 
				  PRINT_STRING_DEFAULT );
	}
	
if ( !( IS_USE_NA( model_P->pullupRef.typ ) ) )
	{
	Get_TypMinMax_as_LString(model_P->pullupRef, s, IBIS_PRINT_WIDTH-8, "V");
  	fprintString( outFile, "[Pullup Reference]        ", IBIS_PRINT_WIDTH, s, 
				  PRINT_STRING_DEFAULT );
	}
	
if ( !( IS_USE_NA( model_P->pulldownRef.typ ) ) )
	{
	Get_TypMinMax_as_LString(model_P->pulldownRef, s, IBIS_PRINT_WIDTH-8, "V");
  	fprintString( outFile, "[Pulldown Reference]      ", IBIS_PRINT_WIDTH, s, 
				  PRINT_STRING_DEFAULT );
	}
	
if ( !( IS_USE_NA( model_P->powerClampRef.typ ) ) )
	{
	Get_TypMinMax_as_LString(model_P->powerClampRef, s, IBIS_PRINT_WIDTH-8, "V");
  	fprintString( outFile, "[POWER Clamp Reference]   ", IBIS_PRINT_WIDTH, s, 
				  PRINT_STRING_DEFAULT );
	}
	
if ( !( IS_USE_NA( model_P->gndClampRef.typ ) ) )
	{
	Get_TypMinMax_as_LString(model_P->gndClampRef, s, IBIS_PRINT_WIDTH-8, "V");
  	fprintString( outFile, "[GND Clamp Reference]     ", IBIS_PRINT_WIDTH, s, 
				  PRINT_STRING_DEFAULT );
	}

if ( model_P->pulldown != NULL )
	{	
	fprintString( outFile, "[Pulldown]", IBIS_PRINT_WIDTH, "", 
				  PRINT_STRING_DEFAULT);
	print_VItable( outFile, model_P->pulldown );
	}

if ( model_P->pullup != NULL )
	{
	fprintString( outFile, "[Pullup]", IBIS_PRINT_WIDTH, "",
				  PRINT_STRING_DEFAULT);
	print_VItable( outFile, model_P->pullup );
	}

if ( model_P->gnd_clamp != NULL )
	{
	fprintString( outFile, "[GND_clamp]", IBIS_PRINT_WIDTH, "", 
				  PRINT_STRING_DEFAULT);
	print_VItable( outFile, model_P->gnd_clamp );
	}

if ( model_P->power_clamp != NULL )
	{
	fprintString( outFile, "[POWER_clamp]", IBIS_PRINT_WIDTH, "", 
				  PRINT_STRING_DEFAULT);
	print_VItable( outFile, model_P->power_clamp);
	}

if ( !( IS_USE_NA( model_P->ramp.dv_r.typ ) ) )
	print_Ramp( outFile, &( model_P->ramp ) );

risingWave_P = model_P->risingWave_P;
fallingWave_P = model_P->fallingWave_P;

while ( risingWave_P != NULL )
	{
	fprintString( outFile, "[Rising Waveform]", IBIS_PRINT_WIDTH, "",
				  PRINT_STRING_DEFAULT );
	Get_Double_as_LString( risingWave_P->R_fixture, numberBuffer, "" );
	fprintString( outFile, "R_fixture = ", IBIS_PRINT_WIDTH, numberBuffer,
				  PRINT_STRING_DEFAULT );
	Get_Double_as_LString( risingWave_P->V_fixture, numberBuffer, "" );
	fprintString( outFile, "V_fixture = ", IBIS_PRINT_WIDTH, numberBuffer,
				  PRINT_STRING_DEFAULT );
	if ( !( IS_USE_NA( risingWave_P->V_fixture_min ) ) )
		{
		Get_Double_as_LString( risingWave_P->V_fixture_min, numberBuffer, "" );
		fprintString( outFile, "V_fixture_min = ", IBIS_PRINT_WIDTH, 
					  numberBuffer, PRINT_STRING_DEFAULT );
		}
	if ( !( IS_USE_NA( risingWave_P->V_fixture_max ) ) )
		{
		Get_Double_as_LString( risingWave_P->V_fixture_max, numberBuffer, "" );
		fprintString( outFile, "V_fixture_max = ", IBIS_PRINT_WIDTH, 
					  numberBuffer, PRINT_STRING_DEFAULT );
		}
	if ( !( IS_USE_NA( risingWave_P->L_fixture ) ) )
		{
		Get_Double_as_LString( risingWave_P->L_fixture, numberBuffer, "H" );
		fprintString( outFile, "L_fixture = ", IBIS_PRINT_WIDTH, 
					  numberBuffer, PRINT_STRING_DEFAULT );
		}
	if ( !( IS_USE_NA( risingWave_P->C_fixture ) ) )
		{
		Get_Double_as_LString( risingWave_P->C_fixture, numberBuffer, "F" );
		fprintString( outFile, "C_fixture = ", IBIS_PRINT_WIDTH, 
					  numberBuffer, PRINT_STRING_DEFAULT );
		}
	if ( !( IS_USE_NA( risingWave_P->R_dut ) ) )
		{
		Get_Double_as_LString( risingWave_P->R_dut, numberBuffer, "" );
		fprintString( outFile, "R_dut = ", IBIS_PRINT_WIDTH, 
					  numberBuffer, PRINT_STRING_DEFAULT );
		}
	if ( !( IS_USE_NA( risingWave_P->L_dut ) ) )
		{
		Get_Double_as_LString( risingWave_P->L_dut, numberBuffer, "H" );
		fprintString( outFile, "L_dut = ", IBIS_PRINT_WIDTH, 
					  numberBuffer, PRINT_STRING_DEFAULT );
		}
	if ( !( IS_USE_NA( risingWave_P->C_dut ) ) )
		{
		Get_Double_as_LString( risingWave_P->C_dut, numberBuffer, "F" );
		fprintString( outFile, "C_dut = ", IBIS_PRINT_WIDTH, 
					  numberBuffer, PRINT_STRING_DEFAULT );
		}
	printWaveTable( outFile, risingWave_P );

	risingWave_P = risingWave_P->nextWaveTable_P;
	}

while ( fallingWave_P != NULL )
	{
	fprintString( outFile, "[Falling Waveform]", IBIS_PRINT_WIDTH, "",
				  PRINT_STRING_DEFAULT );
	Get_Double_as_LString( fallingWave_P->R_fixture, numberBuffer, "" );
	fprintString( outFile, "R_fixture = ", IBIS_PRINT_WIDTH, numberBuffer,
				  PRINT_STRING_DEFAULT );
	Get_Double_as_LString( fallingWave_P->V_fixture, numberBuffer, "" );
	fprintString( outFile, "V_fixture = ", IBIS_PRINT_WIDTH, numberBuffer,
				  PRINT_STRING_DEFAULT );
	if ( !( IS_USE_NA( fallingWave_P->V_fixture_min ) ) )
		{
		Get_Double_as_LString( fallingWave_P->V_fixture_min, numberBuffer, "" );
		fprintString( outFile, "V_fixture_min = ", IBIS_PRINT_WIDTH, 
					  numberBuffer, PRINT_STRING_DEFAULT );
		}
	if ( !( IS_USE_NA( fallingWave_P->V_fixture_max ) ) )
		{
		Get_Double_as_LString( fallingWave_P->V_fixture_max, numberBuffer, "" );
		fprintString( outFile, "V_fixture_max = ", IBIS_PRINT_WIDTH, 
					  numberBuffer, PRINT_STRING_DEFAULT );
		}
	if ( !( IS_USE_NA( fallingWave_P->L_fixture ) ) )
		{
		Get_Double_as_LString( fallingWave_P->L_fixture, numberBuffer, "H" );
		fprintString( outFile, "L_fixture = ", IBIS_PRINT_WIDTH, 
					  numberBuffer, PRINT_STRING_DEFAULT );
		}
	if ( !( IS_USE_NA( fallingWave_P->C_fixture ) ) )
		{
		Get_Double_as_LString( fallingWave_P->C_fixture, numberBuffer, "F" );
		fprintString( outFile, "C_fixture = ", IBIS_PRINT_WIDTH, 
					  numberBuffer, PRINT_STRING_DEFAULT );
		}
	if ( !( IS_USE_NA( fallingWave_P->R_dut ) ) )
		{
		Get_Double_as_LString( fallingWave_P->R_dut, numberBuffer, "" );
		fprintString( outFile, "R_dut = ", IBIS_PRINT_WIDTH, 
					  numberBuffer, PRINT_STRING_DEFAULT );
		}
	if ( !( IS_USE_NA( fallingWave_P->L_dut ) ) )
		{
		Get_Double_as_LString( fallingWave_P->L_dut, numberBuffer, "H" );
		fprintString( outFile, "L_dut = ", IBIS_PRINT_WIDTH, 
					  numberBuffer, PRINT_STRING_DEFAULT );
		}
	if ( !( IS_USE_NA( fallingWave_P->C_dut ) ) )
		{
		Get_Double_as_LString( fallingWave_P->C_dut, numberBuffer, "F" );
		fprintString( outFile, "C_dut = ", IBIS_PRINT_WIDTH, 
					  numberBuffer, PRINT_STRING_DEFAULT );
		}
	printWaveTable( outFile, fallingWave_P );

	fallingWave_P = fallingWave_P->nextWaveTable_P;
	}

sprintf( s, "| End [Model] %s\n|", model_P->modelName );
fprintString( outFile, "", IBIS_PRINT_WIDTH, s, PRINT_INDENT_ONCE );
}

/*
 * delete_Model
 *
 * Delete model data structure
 */
delete_Model(IN ibisModel_Pt model_P)  /* This is the pointer
                                          to the model to delete */
{
delete_VItableEntries(model_P->pulldown);
delete_VItableEntries(model_P->pullup);
delete_VItableEntries(model_P->gnd_clamp);
delete_VItableEntries(model_P->power_clamp);
return;
}



/* Pin Objects */

/* Function ***************************************************************
*
* NAME : getMatchingPin
*
* INPUTS :  char   		*searchName -- pointer to pin name to find
*			ibisPin_Pt	pinList_P -- pointer to pin list to search
*
*
* OUTPUTS :  ibisPin_Pt -- points to pin with name searchName
*
*
* NOTES / SIDE EFFECTS : Finds pin with name searchName.  Returns null if pin
*						 does not exist.
*
*F*/

ibisPin_Pt
getMatchingPin(	IN	char		*searchName,
				IN	ibisPin_Pt	pinList_P )

{
ibisPin_Pt	tempPin_P = pinList_P;

/*------------------------------------------------------------------------
*
*	If the search name is the null string (as may happen with pins with 
*	no input pin), just return the null pointer for the result.
*
*-----------------------------------------------------------------------*/
if ( !( strcmp( searchName, "" ) ) )
	return( NULL );

/*------------------------------------------------------------------------
*
*	Otherwise, search through the pin list until you find the matching
*	pin.
*
*-----------------------------------------------------------------------*/
while( ( tempPin_P != NULL )  && strcmp( searchName, tempPin_P->pinName ) )
	tempPin_P = tempPin_P->nextPin_P;

/*------------------------------------------------------------------------
*
*	If you reach the end of the list without finding a match, report it.
*
*-----------------------------------------------------------------------*/
if ( tempPin_P == NULL )
	{
	fprintf( stderr, "s2ibis2: pin %s not found in pin list.\n",
			 searchName );
	}

return( tempPin_P );
}


/* Function ***************************************************************
*
* NAME : copyPinMapToPins
*
* INPUTS :	ibisComponent_Pt	currentComponent_P	ptr to component
*
*
* OUTPUTS :     0 if successful
*               1 if error 
*
*
* RETURNS : ibisComponent_Pt	currentComponent_P	ptr to component
*
*
* NOTES / SIDE EFFECTS :	This routine copies the data found in the
*							pin map to the corresponding pin.
*F*/

int
copyPinMapToPins( INOUT ibisComponent_Pt	currentComponent_P )

{
ibisPinMap_Pt	currentPinMap_P;
ibisPin_Pt		currentPin_P;

for ( currentPin_P = currentComponent_P->pins_P;
	  currentPin_P != NULL;
	  currentPin_P = currentPin_P->nextPin_P )
	{
	currentPinMap_P = getMatchingPinMap( currentPin_P->pinName, 
							 			 currentComponent_P->pinMap_P );
	if ( currentPinMap_P == NULL )
		return( 1 );
	strcpy( currentPin_P->pullupRef, currentPinMap_P->pullupRef );
	strcpy( currentPin_P->pulldownRef, currentPinMap_P->pulldownRef );
	strcpy( currentPin_P->powerClampRef, currentPinMap_P->powerClampRef );
	strcpy( currentPin_P->gndClampRef, currentPinMap_P->gndClampRef );
	}
return( 0 );
}


/* Function ***************************************************************
*
* NAME : linkPinsToModels
*
* INPUTS :	ibisComponent_Pt	currentComponent_P	ptr to component
*
*
* OUTPUTS :     0 if successful
*               1 if error 
*
*
* RETURNS : ibisComponent_Pt	currentComponent_P	ptr to component
*
*
* NOTES / SIDE EFFECTS :	This routine links each pin to its corresponding
*							model, unless the model name is POWER, GND or NC.
*F*/

int
linkPinsToModels( INOUT	ibisComponent_Pt	currentComponent_P )

{
ibisModel_Pt	currentModel_P;
ibisPin_Pt		currentPin_P;

for ( currentPin_P = currentComponent_P->pins_P;
	  currentPin_P != NULL;
	  currentPin_P = currentPin_P->nextPin_P )
	{
	if ( strcmp( currentPin_P->modelName, "POWER" ) &&
		 strcmp( currentPin_P->modelName, "GND" ) &&
		 strcmp( currentPin_P->modelName, "NC" ) )
		{
		currentModel_P = getMatchingModel( currentPin_P->modelName,
										   currentComponent_P->models_P );
		if ( currentModel_P == NULL )
			{
			fprintf(stderr,"         Please check that you have " );
			fprintf(stderr,"correctly spelled the model\n" );
			fprintf(stderr,"         name for pin %s in the ", 
					currentPin_P->pinName );
			fprintf(stderr,"[Pin] section of your input\n" );
			fprintf(stderr,"         file, and that you have " );
			fprintf(stderr,"specified the model using the\n" );
			fprintf(stderr,"         [Model] keyword before the " );
			fprintf(stderr,"end of\n" );
			fprintf(stderr,"         component %s.\n", 
					currentComponent_P->component );
			return( 1 );
			}
		currentPin_P->model_P = currentModel_P;
		}
	}
return( 0 );
}


/*
 * create_Pin
 *
 * Create pin data structure and NULL all entries.
 */
create_Pin(IN  ibisPin_Pt *top_Pointer_PP, /* This must already exist
                                            * and *top_Pointer_P must point
                                            * to the first pin or a
                                            * subsequent pin in the linked
                                            * list.  */
           OUT ibisPin_Pt *pin_PP)         /* On exit this is the pointer to
                                            *the new pin_P created. */
{
ibisPin_Pt	temp_P;

NEW_LINK( top_Pointer_PP, pin_PP, nextPin_P, ibisPin_Pt, ibisPin_t)
if ( *pin_PP != NULL )
	{
	temp_P = *pin_PP;
	temp_P->R_pin = NOT_USED;
	temp_P->L_pin = NOT_USED;
	temp_P->C_pin = NOT_USED;
	}

return( *pin_PP == NULL);
}


/*
 * print_Pin_Head
 *
 * Print pin data structure.
 */
print_Pin_Head( IN FILE *outFile )
{
char s[MAX_IBIS_STRING_LENGTH];

sprintf( s, "[Pin]" );
fprintString( outFile, "", IBIS_PRINT_WIDTH,
   "[Pin]  signal_name          model_name           R_pin     L_pin     C_pin",
   PRINT_STRING_DEFAULT);

}

/*
 * print_Pin
 *
 * Print pin data structure.
 */
print_Pin(	IN	FILE		*outFile, 
			IN  ibisPin_Pt 	pin_P)

{
char s[MAX_IBIS_STRING_LENGTH];
char tempBuf[MAX_R_PIN_STRING_LENGTH + 1];

sprintf( s, "%s", pin_P->pinName );
spaceFill( s, 7 - strlen( s ) );
strcat( s, pin_P->signalName );
spaceFill( s, 28 - strlen( s ) );
strcat( s, pin_P->modelName );
if ( !( IS_NOT_USED( pin_P->R_pin ) ) )
	{
	spaceFill( s, 49 - strlen( s ) );
	Get_Double_as_LString( pin_P->R_pin, tempBuf, "" );
	strcat( s, tempBuf ); 
	spaceFill( s, 59 - strlen( s ) );
	Get_Double_as_LString( pin_P->L_pin, tempBuf, "H" );
	strcat( s, tempBuf ); 
	spaceFill( s, 69 - strlen( s ) );
	Get_Double_as_LString( pin_P->C_pin, tempBuf, "F" );
	strcat( s, tempBuf ); 
	}
fprintString( outFile, "", IBIS_PRINT_WIDTH, s, PRINT_STRING_DEFAULT );
}

/*
 * delete_Pin
 *
 * Delete pin data structure
 */
delete_Pin(IN ibisPin_Pt pin_P)  /* This is the pointer to the pin to delete */
{
return;
}




/* Ramp Objects */

/*
 * print_Ramp
 *
 * print ramp data structure.  The ramp structure contains separate
 * dv and dt parts, so we have to handle this carefully.
 */
print_Ramp(	IN	FILE		*outFile,
			IN  ibisRamp_Pt ramp_P)
{
char s[MAX_IBIS_STRING_LENGTH];
char dummyBuffer[128];
char numberBuffer[16];

s[0] = 0;
dummyBuffer[0] = 0;
numberBuffer[0] = 0;
fprintString( outFile, "[Ramp]", IBIS_PRINT_WIDTH, "", PRINT_INDENT_ONCE);

printGenericTypMinMaxHeader( outFile );

if ( !( IS_USE_NA( ramp_P->dv_r.typ ) ) )
	{
	Get_Double_as_LString( ramp_P->dv_r.typ, s, "");
	Get_Double_as_LString( ramp_P->dt_r.typ, numberBuffer, "" );
	sprintf( dummyBuffer, "/%s", numberBuffer );
	strcat( s, dummyBuffer );
	}
else
	sprintf( s, "NA" );
spaceFill( s, 20 - strlen( s ) );

if ( !( IS_USE_NA( ramp_P->dv_r.min ) ) )
	{
	Get_Double_as_LString( ramp_P->dv_r.min, numberBuffer, "");
	strcat( s, numberBuffer );
	Get_Double_as_LString( ramp_P->dt_r.min, numberBuffer, "" );
	sprintf( dummyBuffer, "/%s", numberBuffer );
	strcat( s, dummyBuffer );
	}
else
	{
	sprintf( dummyBuffer, "NA" );
	strcat( s, dummyBuffer );
	}
spaceFill( s, 40 - strlen( s ) );

if ( !( IS_USE_NA( ramp_P->dv_r.max ) ) )
	{
	Get_Double_as_LString( ramp_P->dv_r.max, numberBuffer, "");
	strcat( s, numberBuffer );
	Get_Double_as_LString( ramp_P->dt_r.max, numberBuffer, "" );
	sprintf( dummyBuffer, "/%s", numberBuffer );
	strcat( s, dummyBuffer );
	}
else
	{
	sprintf( dummyBuffer, "NA" );
	strcat( s, dummyBuffer );
	}

fprintString( outFile, "dV/dt_r          ", IBIS_PRINT_WIDTH, s, 
			  PRINT_STRING_DEFAULT);

if ( !( IS_USE_NA( ramp_P->dv_f.typ ) ) )
	{
	Get_Double_as_LString( ramp_P->dv_f.typ, s, "");
	Get_Double_as_LString( ramp_P->dt_f.typ, numberBuffer, "" );
	sprintf( dummyBuffer, "/%s", numberBuffer );
	strcat( s, dummyBuffer );
	}
else
	sprintf( s, "NA" );
spaceFill( s, 20 - strlen( s ) );

if ( !( IS_USE_NA( ramp_P->dv_f.min ) ) ) 
	{
	Get_Double_as_LString( ramp_P->dv_f.min, numberBuffer, "");
	strcat( s, numberBuffer );
	Get_Double_as_LString( ramp_P->dt_f.min, numberBuffer, "" );
	sprintf( dummyBuffer, "/%s", numberBuffer );
	strcat( s, dummyBuffer );
	}
else
	{
	sprintf( dummyBuffer, "NA" );
	strcat( s, dummyBuffer );
	}
spaceFill( s, 40 - strlen( s ) );

if ( !( IS_USE_NA( ramp_P->dv_f.max ) ) )
	{
	Get_Double_as_LString( ramp_P->dv_f.max, numberBuffer, "");
	strcat( s, numberBuffer );
	Get_Double_as_LString( ramp_P->dt_f.max, numberBuffer, "" );
	sprintf( dummyBuffer, "/%s", numberBuffer );
	strcat( s, dummyBuffer );
	}
else
	{
	sprintf( dummyBuffer, "NA" );
	strcat( s, dummyBuffer );
	}
fprintString( outFile, "dV/dt_f          ", IBIS_PRINT_WIDTH, s, 
			  PRINT_STRING_DEFAULT);

fprintString( outFile, "|", IBIS_PRINT_WIDTH, "", PRINT_INDENT_ONCE );
}




/* VITable Objects */

/*
 * create_VItable
 *
 * Create a VI table of size size
 */
create_VItable(ibisVItable_Pt *VItable_PP, int size)
  {
  if( (*VItable_PP = (ibisVItable_Pt) MALLOC(sizeof(ibisVItable_t) ) ) == NULL)
    {
    yyerror("Out of memory");
    return(1);
    }
  return(resize_VItable(*VItable_PP, size));
  }

/*
 * resize_VItable
 */
resize_VItable(INOUT ibisVItable_Pt VItable_P, IN int size)
{
ibisVItableEntry_Pt newVIs_P;
int newSize, i;
if( size == 0)
  newSize = MAX_TABLE_SIZE; /* Use default table size */
else
  newSize = size;
if( ( newVIs_P = (ibisVItableEntry_Pt)
                   MALLOC(sizeof(ibisVItableEntry_t) * newSize) ) == NULL)
  {
  yyerror("Out of memory");
  return(1);
  }
/* To make life easier, fill the new table with USE_NA */
for ( i = 0; i < newSize; i++ )
	{
	newVIs_P[i].i.typ = USE_NA;
	newVIs_P[i].i.min = USE_NA;
	newVIs_P[i].i.max = USE_NA;
	}

/* If there was no previous table, just point at the new one */
if(VItable_P->VIs ==  NULL)
  {
  VItable_P->VIs = newVIs_P;
  VItable_P->size = newSize;
  }
else				/* otherwise, copy all the old data values in */
  {
  for(i = 0; i < VItable_P->size; i++)
    {
    newVIs_P[i] = VItable_P->VIs[i];
    }
  delete_VItableEntries(VItable_P);
  VItable_P->VIs = newVIs_P;
  VItable_P->size = newSize;
  }
return(0);
}


/*
 * duplicate_VItable
 *
 * Create a VI table destVItable of size size and copy entries
 * from sourceVItable
 */
duplicate_VItable(IN  ibisVItable_Pt sourceVItable_P,
               OUT ibisVItable_Pt *destVItable_PP, IN int size)
{
ibisVItableEntry_Pt newVIs_P, oldVIs_P;

create_VItable(destVItable_PP, size);
newVIs_P = &((*destVItable_PP)->VIs[0]);
oldVIs_P = &(sourceVItable_P->VIs[0]);
for(; size; size--)
  *newVIs_P++ = *oldVIs_P++;
}


/*
 * print_VItable
 *
 * Print VITable data structure
 */
print_VItable(	IN	FILE			*outFile,
				IN	ibisVItable_Pt 	VItable_P )
{
int size = VItable_P->size, i;
char s1[30];
char s[MAX_IBIS_STRING_LENGTH];

if ( VItable_P == NULL ) 
	return(1);
if ( VItable_P->VIs == NULL ) 
	return(1);

Get_TypMinMax_header_as_String(s);
fprintString( outFile, "| voltage   ", IBIS_PRINT_WIDTH, s, PRINT_INDENT_ONCE);
fprintString( outFile, "|", IBIS_PRINT_WIDTH, "", PRINT_INDENT_ONCE);

for(i = 0; i < size; i++)
  {
  sprintf(s1,"%7.2f     ",VItable_P->VIs[i].v);
  Get_TypMinMax_as_String(VItable_P->VIs[i].i, s, IBIS_PRINT_WIDTH-8, "A");
  fprintString( outFile, s1, IBIS_PRINT_WIDTH, s, PRINT_INDENT_ONCE);
  }
fprintString( outFile, "|", IBIS_PRINT_WIDTH, "", PRINT_INDENT_ONCE );
return(0);
}

/*
 * delete_VItableEntries
 *
 * Destroy a VITable data structure
 */
delete_VItableEntries(ibisVItable_Pt VItable_P)
{
FREE(VItable_P->VIs);
}


/*
 * delete_VItable
 *
 * Destroy a VITable data structure
 */
delete_VItable(ibisVItable_Pt *VItable_PP)
{
delete_VItableEntries(*VItable_PP);
FREE(*VItable_PP);
}




/* WaveTable Objects */


/* Function ***************************************************************
*
* NAME : createWaveTable
*
* INPUTS : 	waveTable_PP -- handle to wave table structure
*
*
* OUTPUTS :	waveTable_PP -- handle to wave table structure
*
*
* RETURNS :	status of resizeWaveTable if initial allocation successful
*		1 if initial allocation fails
*
*
* NOTES / SIDE EFFECTS : 
*
*
*F*/

createWaveTable(ibisWaveTable_Pt *waveTable_PP, int size)
  {
  if( (*waveTable_PP = (ibisWaveTable_Pt) MALLOC(sizeof(ibisWaveTable_t) ) ) == NULL)
    {
    yyerror("Out of memory");
    return(1);
    }
  return(resizeWaveTable(*waveTable_PP, size));
  }


/* Function ***************************************************************
*
* NAME : resizeWaveTable
*
* INPUTS :	ibisWaveTable_Pt waveTable_P -- pointer to wave table structure
*			int size -- desired size of wave table 
*
*
* OUTPUTS : ibisWaveTable_Pt waveTable_P -- pointer to wave table structure
*
*
* RETURNS : 	0 if successful
*		1 if insufficient memory
*
*
* NOTES / SIDE EFFECTS : 
*
*
*F*/

resizeWaveTable( INOUT ibisWaveTable_Pt waveTable_P,
		 IN int size)
{


/*==== Declarations =====================================================*/
ibisWaveTableEntry_Pt	newWave_P;		/* temp pointer to data */
int			newSize, i;

/*==== Code =============================================================*/
if ( size==0 )					/* If no size info... */
	newSize = MAX_TABLE_SIZE; /* ...use default */
else
	newSize = size;

if ( (newWave_P = (ibisWaveTableEntry_Pt)	/* Allocate memory for... */
      MALLOC(sizeof(ibisWaveTableEntry_t) * newSize) ) == NULL)
	{										/* ...new table; fail...*/
	yyerror("Out of memory");				/* ...if insuff. mem. */
	return(1);
	}
/* To make life easier, fill the new table with USE_NA */
for ( i = 0; i < newSize; i++ )
	{
	newWave_P[i].v.typ = USE_NA;
	newWave_P[i].v.min = USE_NA;
	newWave_P[i].v.max = USE_NA;
	}
if ( waveTable_P->waveData == NULL )		/* If no previous wave...  */
	{					/* ...data, just assign... */
	waveTable_P->waveData = newWave_P;	/* ...the space. */
	waveTable_P->size = newSize;
	}
else						/* Otherwise... */
	{					/* ...copy the old data... */
	for ( i=0; i < waveTable_P->size; i++ )
		newWave_P[i] = waveTable_P->waveData[i];
	deleteWaveTableEntries( waveTable_P );	/* ...delete the old... */
	waveTable_P->waveData = newWave_P;	/* ...and point to the new */
	waveTable_P->size = newSize;
	}
return(0);
}

/* Function ***************************************************************
*
* NAME : duplicateWaveTable
*
* INPUTS :	size -- number of entries in source table
*		sourceWaveTable_P -- pointer to source wave table
*
*
* OUTPUTS : 	destWaveTable_P -- pointer to destination wave table
*
*
* RETURNS :	none
*
*
* NOTES / SIDE EFFECTS : 
*
*
*F*/

duplicateWaveTable(	IN ibisWaveTable_Pt sourceWaveTable_P,
			OUT ibisWaveTable_Pt *destWaveTable_PP, 
			IN int size)
{

/*==== Declarations =====================================================*/
ibisWaveTableEntry_Pt	newWave_P, 	/* temp pointer to new data */
			oldWave_P;	/* temp pointer to old data */


/*==== Code =============================================================*/
createWaveTable( destWaveTable_PP, size );		/* Create table */
newWave_P = &( ( *destWaveTable_PP )->waveData[0] );	/* Point to new */
oldWave_P = &( sourceWaveTable_P->waveData[0] );     	/* Point to old */

for ( ; size; size-- )		       /* Copy data points one by one. */
	*newWave_P++ = *oldWave_P++;
}



/* Function ***************************************************************
*
* NAME : printWaveTable
*
* INPUTS : 	FILE				*outFile
*			ibisWaveTable_Pt	waveTable_P
*
*
* OUTPUTS : 0 on successful completion, 
*	    1 if pointer is null or if waveData pointer is null.
*
*F*/


printWaveTable(	IN	FILE				*outFile,
				IN	ibisWaveTable_Pt 	waveTable_P )
{

/*==== Declarations =====================================================*/
int	size = waveTable_P->size;
int	i;
char	s1[30];
char	s[MAX_IBIS_STRING_LENGTH];

/*==== Code =============================================================*/
if ( waveTable_P == NULL ) return(1);		/* If it's empty... */
						/* ...or no data... */
if ( waveTable_P->waveData == NULL ) return(1);	/* ...return a 1. */


					/* Otherwise, print a header line */
Get_VTypMinMax_header_as_String( s );
fprintString( outFile, "| time         ", IBIS_PRINT_WIDTH, s, 
			  PRINT_INDENT_ONCE );
fprintString( outFile, "|", IBIS_PRINT_WIDTH, "", PRINT_INDENT_ONCE );

					/* Then print the data points... */
for ( i = 0; i < size; i++ )		/* ...one line at a time. */
	{
	Get_Double_as_String( waveTable_P->waveData[i].t, s1, "S" );
	spaceFill( s1, 15 - strlen( s1 ) );
	Get_TypMinMax_as_String( waveTable_P->waveData[i].v, s, 
				 IBIS_PRINT_WIDTH - 8, "V" );
	fprintString( outFile, s1, IBIS_PRINT_WIDTH, s, PRINT_INDENT_ONCE );
	}
fprintString( outFile, "|", IBIS_PRINT_WIDTH, "", PRINT_INDENT_ONCE );
return(0);
}


/* Function ***************************************************************
*
* NAME : deleteWaveTableEntries
*
* INPUTS : waveTable_P -- pointer to wave table structure
*
*
* NOTES / SIDE EFFECTS : deletes a single wave tabel entry.
*
*
*F*/

deleteWaveTableEntries( ibisWaveTable_Pt	waveTable_P )
{

/*==== Code =============================================================*/
FREE( waveTable_P->waveData );
}



/* Function ***************************************************************
*
* NAME : deleteWaveTable
*
* INPUTS : waveTable_PP -- handle to wave table
*
*
* NOTES / SIDE EFFECTS : deletes the whole wave table
*
*
*F*/

deleteWaveTable( ibisWaveTable_Pt 	*waveTable_PP )
{

/*==== Code =============================================================*/
deleteWaveTableEntries( *waveTable_PP );
FREE( *waveTable_PP );
}



/* Component Object */

/*
 * create_Component
 *
 * Create component data structure and NULL all entries.
 * Insert at end of linked list.
 */
create_Component( 	IN	ibisComponent_Pt *top_Pointer_PP,
                          	/* This must already exist and *top_Pointer_PP must
                          	 * point to the first comp in the linked list.*/
             		OUT ibisComponent_Pt *component_PP )       
							/* This is the pointer to the new entry. */
{
ibisComponent_Pt	temp_P;

NEW_LINK( top_Pointer_PP, component_PP, nextComponent_P,
           ibisComponent_Pt, ibisComponent_t );
if ( *component_PP != NULL )
	{
	temp_P = *component_PP;
	temp_P->tempRange.typ = USE_NA;
	temp_P->tempRange.min = USE_NA;
	temp_P->tempRange.max = USE_NA;
	temp_P->voltageRange.typ = USE_NA;
	temp_P->voltageRange.min = USE_NA;
	temp_P->voltageRange.max = USE_NA;
	temp_P->pullupRef.typ = USE_NA;
	temp_P->pullupRef.min = USE_NA;
	temp_P->pullupRef.max = USE_NA;
	temp_P->pulldownRef.typ = USE_NA;
	temp_P->pulldownRef.min = USE_NA;
	temp_P->pulldownRef.max = USE_NA;
	temp_P->powerClampRef.typ = USE_NA;
	temp_P->powerClampRef.min = USE_NA;
	temp_P->powerClampRef.max = USE_NA;
	temp_P->gndClampRef.typ = USE_NA;
	temp_P->gndClampRef.min = USE_NA;
	temp_P->gndClampRef.max = USE_NA;
	temp_P->c_comp.typ = USE_NA;
	temp_P->c_comp.min = USE_NA;
	temp_P->c_comp.max = USE_NA;

	temp_P->pinParasitics.R_pkg.typ = USE_NA;
	temp_P->pinParasitics.R_pkg.min = USE_NA;
	temp_P->pinParasitics.R_pkg.max = USE_NA;
	temp_P->pinParasitics.L_pkg.typ = USE_NA;
	temp_P->pinParasitics.L_pkg.min = USE_NA;
	temp_P->pinParasitics.L_pkg.max = USE_NA;
	temp_P->pinParasitics.C_pkg.typ = USE_NA;
	temp_P->pinParasitics.C_pkg.min = USE_NA;
	temp_P->pinParasitics.C_pkg.max = USE_NA;

	temp_P->Rload = USE_NA;
	temp_P->derateVIPct = USE_NA;
	temp_P->derateRampPct = USE_NA;
	}
		   
return( *component_PP == NULL );
}



void
printComponentHead(	IN	FILE	*outFile,
					IN	char	*componentName )

{
char	tempBuffer[IBIS_PRINT_WIDTH];
int		spacing;

nullString( tempBuffer, IBIS_PRINT_WIDTH );
fprintString( outFile, "|", IBIS_PRINT_WIDTH, 
	"************************************************************************",
	PRINT_INDENT_ONCE );
spacing = ( 72 - ( strlen( componentName ) + strlen( "Component " ) ) ) / 2 ;
spaceFill( tempBuffer, spacing );
strcat( tempBuffer, "Component " );
strcat( tempBuffer, componentName );
fprintString( outFile, "|", IBIS_PRINT_WIDTH, tempBuffer, PRINT_INDENT_ONCE );
fprintString( outFile, "|", IBIS_PRINT_WIDTH, 
	"************************************************************************",
	PRINT_INDENT_ONCE );
fprintString( outFile, "|", IBIS_PRINT_WIDTH, "", PRINT_INDENT_ONCE );
}



/*
 * print_IBIS_Version;
 */
print_IBIS_Version(	IN	FILE	*outFile,
					IN 	int 	ibisVersion)
{
switch(ibisVersion)
  {
  case VERSION_ONE_ZERO:
    fprintString( outFile, "[IBIS ver]       ", IBIS_PRINT_WIDTH, "1.0",
				  PRINT_INDENT_ONCE );
    break;
  case VERSION_ONE_ONE:
    fprintString( outFile, "[IBIS ver]       ", IBIS_PRINT_WIDTH, "1.1",
				  PRINT_INDENT_ONCE );
    break;
  case VERSION_TWO_ZERO:
    fprintString( outFile, "[IBIS ver]       ", IBIS_PRINT_WIDTH, "2.0",
				  PRINT_INDENT_ONCE );
    break;
  case VERSION_TWO_ONE:
    fprintString( outFile, "[IBIS ver]       ", IBIS_PRINT_WIDTH, "2.1",
				  PRINT_INDENT_ONCE );
    break;
  default:
    fprintString( outFile, "[IBIS ver]       ", IBIS_PRINT_WIDTH, "unknown",
				  PRINT_INDENT_ONCE);
    break;
  }
}




/*
 * print_Component
 *
 * Print component data structure.
 */
print_Component(	IN	FILE				*outFile,
					IN  ibisComponent_Pt 	component_P)
{
char 			s[MAX_IBIS_STRING_LENGTH];
ibisPin_Pt		tempPin_P = NULL;
ibisPinMap_Pt	tempPinMap_P = NULL;
ibisModel_Pt	tempModel_P = NULL;
ibisDiffPin_Pt	tempDiffPin_P = NULL;

fprintString( outFile, "|", IBIS_PRINT_WIDTH, "", PRINT_INDENT_ONCE);
printComponentHead( outFile, component_P->component );
fprintString( outFile, "[Component]      ", IBIS_PRINT_WIDTH, 
			  component_P->component, PRINT_INDENT_ONCE);
fprintString( outFile, "[Manufacturer]   ", IBIS_PRINT_WIDTH, 
			  component_P->manufacturer, PRINT_INDENT_ONCE);

/* Package */

fprintString( outFile, "", IBIS_PRINT_WIDTH, "[Package]",PRINT_STRING_DEFAULT);
printGenericTypMinMaxHeader( outFile );

Get_TypMinMax_as_LString( component_P->pinParasitics.R_pkg, s, 
						 IBIS_PRINT_WIDTH-8, "" );
fprintString( outFile, "R_pkg            ", IBIS_PRINT_WIDTH, s, 
			  PRINT_INDENT_ONCE);

Get_TypMinMax_as_LString( component_P->pinParasitics.L_pkg, s, 
						 IBIS_PRINT_WIDTH-8, "H" );
fprintString( outFile, "L_pkg            ", IBIS_PRINT_WIDTH, s, 
			  PRINT_INDENT_ONCE);

Get_TypMinMax_as_LString( component_P->pinParasitics.C_pkg, s, 
						 IBIS_PRINT_WIDTH-8, "F" );
fprintString( outFile, "C_pkg            ", IBIS_PRINT_WIDTH, s, 
			  PRINT_INDENT_ONCE);
fprintString( outFile, "|", IBIS_PRINT_WIDTH, "", PRINT_INDENT_ONCE );

/*-------------------------------------------------------------------------
*
*   Print the pin list.
*
*------------------------------------------------------------------------*/
if ( component_P->pins_P != NULL )
	{
	print_Pin_Head( outFile );
	for ( tempPin_P = component_P->pins_P;
		  tempPin_P != NULL;
		  tempPin_P = tempPin_P->nextPin_P )
		print_Pin( outFile, tempPin_P );
	fprintString( outFile, "|", IBIS_PRINT_WIDTH, "", PRINT_INDENT_ONCE );
	}

/*-------------------------------------------------------------------------
*
*   If there is a package model, print it.
*
*------------------------------------------------------------------------*/
if ( strcmp( component_P->packageModel, "" ) )
	{
	fprintString( outFile, "[Package Model]  ", IBIS_PRINT_WIDTH, 
				  component_P->packageModel, PRINT_INDENT_ONCE );
	fprintString( outFile, "|", IBIS_PRINT_WIDTH, "", PRINT_INDENT_ONCE );
	}

/*-------------------------------------------------------------------------
*
*   If there is a pin map, print it.
*
*------------------------------------------------------------------------*/
if ( component_P->hasPinMapping )
	{
	fprintString( outFile, "", IBIS_PRINT_WIDTH, 
"[Pin Mapping]  pulldown_ref    pullup_ref      gnd_clamp_ref   power_clamp_ref", PRINT_INDENT_ONCE );
	fprintString( outFile, "|", IBIS_PRINT_WIDTH, "", PRINT_INDENT_ONCE );
	for ( tempPinMap_P = component_P->pinMap_P;
		  tempPinMap_P != NULL;
		  tempPinMap_P = tempPinMap_P->nextPinMap_P )
		printPinMap( outFile, tempPinMap_P );
	fprintString( outFile, "|", IBIS_PRINT_WIDTH, "", PRINT_INDENT_ONCE );
	}

/*-------------------------------------------------------------------------
*
*   If there's a diff pin list, print it.
*
*------------------------------------------------------------------------*/
if ( component_P->diffPin_P != NULL )
	{
	fprintString( outFile, "", IBIS_PRINT_WIDTH, 
		"[Diff Pin]  inv_pin  vdiff     tdelay_typ  tdelay_min  tdelay_max",
		PRINT_INDENT_ONCE );
	fprintString( outFile, "|", IBIS_PRINT_WIDTH, "", PRINT_INDENT_ONCE );
	for ( tempDiffPin_P = component_P->diffPin_P;
		  tempDiffPin_P != NULL;
		  tempDiffPin_P = tempDiffPin_P->nextDiffPin_P )
		printDiffPin( outFile, tempDiffPin_P );
	fprintString( outFile, "|", IBIS_PRINT_WIDTH, "", PRINT_INDENT_ONCE );
	}

/*-------------------------------------------------------------------------
*
*   Print all the model results, and that's it.
*
*------------------------------------------------------------------------*/
if ( component_P->models_P != NULL )
	for ( tempModel_P = component_P->models_P;
		  tempModel_P != NULL;
		  tempModel_P = tempModel_P->nextModel_P )
		print_Model( outFile, tempModel_P );

sprintf( s, "| End [Component] %s\n|", component_P->component );
fprintString( outFile, "", IBIS_PRINT_WIDTH, s, PRINT_INDENT_ONCE );
}


/*
 * delete_Component
 *
 * Delete model data structure
 */
delete_Component(IN ibisComponent_Pt component_P)
{
DELETE_LINKED_LIST( &(component_P->pins_P), nextPin_P, delete_Pin, ibisPin_Pt )
DELETE_LINKED_LIST( &(component_P->models_P), nextModel_P, delete_Model, 
					ibisModel_Pt )
DELETE_LINKED_LIST( &(component_P->pinMap_P), nextPinMap_P, deletePinMap,
					ibisPinMap_Pt )
DELETE_LINKED_LIST( &(component_P->diffPin_P), nextDiffPin_P, deleteDiffPin,
					ibisDiffPin_Pt )
return;
}


/* Global Object */

/*
 * createGlobal
 *
 * Create Global ibis data structure and NULL all entries.
 * 
 */
createGlobal( OUT ibisGlobal_Pt *global_P) /* This is the pointer 
											  to the new entry. */
{
ibisGlobal_Pt	temp_P;

if( (*global_P = (ibisGlobal_Pt) MALLOC(sizeof(ibisGlobal_t))) == NULL)
	{
	yyerror("Out of memory");
	return(1);
	}

temp_P = *global_P;
temp_P->tempRange.typ = TEMP_TYP_DEFAULT;
temp_P->tempRange.min = TEMP_MIN_DEFAULT;
temp_P->tempRange.max = TEMP_MAX_DEFAULT;
temp_P->voltageRange.typ = VOLTAGE_RANGE_TYP_DEFAULT;
temp_P->voltageRange.min = VOLTAGE_RANGE_MIN_DEFAULT;
temp_P->voltageRange.max = VOLTAGE_RANGE_MAX_DEFAULT;
temp_P->pullupRef.typ = USE_NA;
temp_P->pullupRef.min = USE_NA;
temp_P->pullupRef.max = USE_NA;
temp_P->pulldownRef.typ = USE_NA;
temp_P->pulldownRef.min = USE_NA;
temp_P->pulldownRef.max = USE_NA;
temp_P->powerClampRef.typ = USE_NA;
temp_P->powerClampRef.min = USE_NA;
temp_P->powerClampRef.max = USE_NA;
temp_P->gndClampRef.typ = USE_NA;
temp_P->gndClampRef.min = USE_NA;
temp_P->gndClampRef.max = USE_NA;
temp_P->c_comp.typ = C_COMP_DEFAULT;
temp_P->c_comp.min = C_COMP_DEFAULT;
temp_P->c_comp.max = C_COMP_DEFAULT;

temp_P->pinParasitics.R_pkg.typ = R_PKG_DEFAULT;
temp_P->pinParasitics.R_pkg.min = R_PKG_DEFAULT;
temp_P->pinParasitics.R_pkg.max = R_PKG_DEFAULT;
temp_P->pinParasitics.L_pkg.typ = L_PKG_DEFAULT;
temp_P->pinParasitics.L_pkg.min = L_PKG_DEFAULT;
temp_P->pinParasitics.L_pkg.max = L_PKG_DEFAULT;
temp_P->pinParasitics.C_pkg.typ = C_PKG_DEFAULT;
temp_P->pinParasitics.C_pkg.min = C_PKG_DEFAULT;
temp_P->pinParasitics.C_pkg.max = C_PKG_DEFAULT;

temp_P->simTime = SIM_TIME_DEFAULT;
temp_P->Rload = RLOAD_DEFAULT;
temp_P->derateVIPct = DERATE_VI_PCT_DEFAULT;
temp_P->derateRampPct = DERATE_RAMP_PCT_DEFAULT;
return( 0 );
}


/* Function ***************************************************************
*
* NAME : copyGlobalDataToComponents
*
* INPUTS :	ibisGlobal_Pt	globals_P	ptr to global data structure
*			ibisTOP_Pt		top_P		ptr to top IBIS structure
*
*
* RETURNS : ibisTOP_Pt		top_P		ptr to top IBIS structure
*
*
* NOTES / SIDE EFFECTS :	This routine copies the global data found in the
*							ibisGlobal_t structure to the components, if
*							it is not already there.
*F*/

void
copyGlobalDataToComponents(	IN		ibisGlobal_Pt	global_P,
							INOUT	ibisTOP_Pt		top_P )

{
ibisComponent_Pt	currentComponent_P;

for ( currentComponent_P = top_P->components_P;
	  currentComponent_P != NULL;
	  currentComponent_P = currentComponent_P->nextComponent_P )
	{
	if ( IS_USE_NA( currentComponent_P->c_comp.typ ) )
		currentComponent_P->c_comp = global_P->c_comp;
	if ( IS_USE_NA( currentComponent_P->pinParasitics.R_pkg.typ ) )
		currentComponent_P->pinParasitics.R_pkg = global_P->pinParasitics.R_pkg;
	if ( IS_USE_NA( currentComponent_P->pinParasitics.L_pkg.typ ) )
		currentComponent_P->pinParasitics.L_pkg = global_P->pinParasitics.L_pkg;
	if ( IS_USE_NA( currentComponent_P->pinParasitics.C_pkg.typ ) )
		currentComponent_P->pinParasitics.C_pkg = global_P->pinParasitics.C_pkg;
	if ( IS_USE_NA( currentComponent_P->tempRange.typ ) )
		currentComponent_P->tempRange = global_P->tempRange;
	if ( IS_USE_NA( currentComponent_P->voltageRange.typ ) &&
		 IS_USE_NA( currentComponent_P->pullupRef.typ ) )
		currentComponent_P->voltageRange = global_P->voltageRange;
	if ( IS_USE_NA( currentComponent_P->pullupRef.typ ) &&
		 IS_USE_NA( currentComponent_P->voltageRange.typ ) )
		currentComponent_P->pullupRef = global_P->pullupRef;
	if ( IS_USE_NA( currentComponent_P->pulldownRef.typ ) &&
		 IS_USE_NA( currentComponent_P->voltageRange.typ ) )
		currentComponent_P->pulldownRef = global_P->pulldownRef;
	if ( IS_USE_NA( currentComponent_P->powerClampRef.typ ) &&
		 IS_USE_NA( currentComponent_P->voltageRange.typ ) )
		currentComponent_P->powerClampRef = global_P->powerClampRef;
	if ( IS_USE_NA( currentComponent_P->gndClampRef.typ ) &&
		 IS_USE_NA( currentComponent_P->voltageRange.typ ) )
		currentComponent_P->gndClampRef = global_P->gndClampRef;
	if ( IS_USE_NA( currentComponent_P->Rload ) )
		currentComponent_P->Rload = global_P->Rload;
	if ( IS_USE_NA( currentComponent_P->derateVIPct ) )
		currentComponent_P->derateVIPct = global_P->derateVIPct;
	if ( IS_USE_NA( currentComponent_P->derateRampPct ) )
		currentComponent_P->derateRampPct = global_P->derateRampPct;
	if ( IS_USE_NA( currentComponent_P->simTime ) )
		currentComponent_P->simTime = global_P->simTime;
	}
}





/* TOP Object */

/*
 * create_TOP
 *
 * Create TOP ibis data structure and NULL all entries.
 * 
 */
create_TOP( OUT ibisTOP_Pt 	*TOP_P, /* This is the pointer to the new entry. */
			IN	char		*file )
{
ibisTOP_Pt	temp_P;
time_t		currentTime;
char		dateBuffer[MAX_DATE_STRING_LENGTH + 1],
			fileNameBuffer[MAX_FILENAME_LENGTH + 1];
int			i = 0;

if( (*TOP_P = (ibisTOP_Pt) MALLOC(sizeof(ibisTOP_t))) == NULL)
	{
	yyerror("Out of memory");
	return(1);
	}

/*-------------------------------------------------------------------------
*
*	Now we need to set all the string pointers to point to the null string.
*	We save about 4 bytes (no, really!) doing it this way.
*
*	We'll also set the date string to the current date.  That will be the
*	default.
*
*	Also, set spiceType to SPICE_TYPE_DEFAULT.
*
*	Finally, we'll default the output filename to the first 8 characters
*	of the input file name plus ".ibs"
*
*------------------------------------------------------------------------*/
temp_P = *TOP_P;
create_String( "", &( temp_P->source ) );
temp_P->notes = temp_P->source;
temp_P->disclaimer = temp_P->source;
temp_P->copyright = temp_P->source;
temp_P->spiceCommand = temp_P->source;

currentTime = time( 0 );
strftime( dateBuffer, MAX_DATE_STRING_LENGTH, "%A, %B %d, %Y", 
		  localtime( &currentTime ) );
LOAD_STRING( temp_P->date, dateBuffer, MAX_DATE_STRING_LENGTH );

temp_P->spiceType = SPICE_TYPE_DEFAULT;

for ( i = 0; i < ( MAX_FILENAME_LENGTH + 1 ); i++ )
	fileNameBuffer[i] = '\0';

i = 0;
while ( ( file[i] != '\0' ) && ( i < MAX_FILENAME_LENGTH + 1 ) ) 
	{
	if ( file[i] == '.' ) 
		file[i] = '\0';
	i++;
	}

strncpy( fileNameBuffer, file, MAX_FILENAME_BASE_LENGTH );
strcat( fileNameBuffer, "." );
strcat( fileNameBuffer, FILENAME_EXTENSION );
LOAD_STRING( temp_P->thisFileName, fileNameBuffer, MAX_FILENAME_LENGTH );
return( 0 );
}


/*
 * print_TOP
 *
 * Print ibis data structure.
 */
print_TOP(	IN	FILE		*outFile,
			IN  ibisTOP_Pt 	TOP_P)
{
ibisComponent_Pt	tempComponent_P = NULL;
char				headerString[IBIS_PRINT_WIDTH];

nullString( headerString, IBIS_PRINT_WIDTH );

sprintf( headerString, " IBIS file %s created by s2ibis2 version %s", 
		 TOP_P->thisFileName, VERSION_STRING );
fprintString( outFile, "|", IBIS_PRINT_WIDTH,
	"************************************************************************",
	PRINT_INDENT_ONCE );
fprintString( outFile, "|", IBIS_PRINT_WIDTH, headerString, PRINT_INDENT_ONCE );
fprintString( outFile, "|", IBIS_PRINT_WIDTH,
	" North Carolina State University Electronics Research Laboratory  1995",
	PRINT_INDENT_ONCE );
fprintString( outFile, "|", IBIS_PRINT_WIDTH,
	"************************************************************************",
	PRINT_INDENT_ONCE );
fprintString( outFile, "|", IBIS_PRINT_WIDTH, "", PRINT_INDENT_ONCE);
print_IBIS_Version( outFile, TOP_P->ibisVersion);
fprintString( outFile, "[File name]      ", IBIS_PRINT_WIDTH, 
			  TOP_P->thisFileName, PRINT_INDENT_ONCE);
fprintString( outFile, "[File Rev]       ", IBIS_PRINT_WIDTH, TOP_P->fileRev,
            PRINT_INDENT_ONCE);
fprintString( outFile, "[Date]           ", IBIS_PRINT_WIDTH, TOP_P->date,
            PRINT_INDENT_ONCE);
fprintString( outFile, "[Source]         ", IBIS_PRINT_WIDTH, TOP_P->source,
            PRINT_INDENT_ONCE);
fprintString( outFile, "[Notes]          ", IBIS_PRINT_WIDTH, TOP_P->notes,
            PRINT_INDENT_ONCE);
fprintString( outFile, "[Disclaimer]     ", IBIS_PRINT_WIDTH, TOP_P->disclaimer,
            PRINT_INDENT_ONCE);
fprintString( outFile, "[Copyright]      ", IBIS_PRINT_WIDTH, TOP_P->copyright,
            PRINT_INDENT_ONCE);

for ( tempComponent_P = TOP_P->components_P;
	  tempComponent_P != NULL;
	  tempComponent_P = tempComponent_P->nextComponent_P )
	print_Component( outFile, tempComponent_P );

fprintString(  outFile, "[End]", IBIS_PRINT_WIDTH, "", PRINT_INDENT_ONCE );

}


/*
 * delete_TOP
 *
 * Delete top data structure
 */
delete_TOP(INOUT  ibisTOP_Pt *TOP_PP)
{
ibisTOP_Pt TOP_P = *TOP_PP;
FREE(TOP_P->source);
FREE(TOP_P->notes);
FREE(TOP_P->disclaimer);
DELETE_LINKED_LIST(&(TOP_P->components_P), nextComponent_P,
                   delete_Component, ibisComponent_Pt )
FREE(TOP_P);
return;
}


/*
 * ibis_Print
 */

void s2ibisPrint(ibisTOP_Pt ibis_P)
{
FILE	*outFile;

fprintf( stderr, "s2ibis2: Opening file %s for writing...", 
		 ibis_P->thisFileName );

if ( ( outFile = fopen( ibis_P->thisFileName, "w" ) ) == NULL )
	{
	fprintf( stderr, "\ns2ibis2: Unable to open output file %s for writing.\n",
			 ibis_P->thisFileName );
	return;
	}
print_TOP( outFile, ibis_P);
fclose( outFile );
fprintf( stderr, "done.\n" );
return;
}



/* Function ***************************************************************
*
* NAME : completeDataStructures
*
* INPUTS : 	ibisGlobal_Pt	global_P	-- ptr to global data structure
*			ibisTOP_Pt		ibis_P		-- ptr to top of ibis structure
*
*
* OUTPUTS :	0 on success, 1 on fail
*
*
* RETURNS : ibisTOP_Pt		ibis_P		-- ptr to top of ibis structure 
*
*	
* NOTES / SIDE EFFECTS : This function propagates global data values down to
*							the component and model level.
*
*F*/

int
completeDataStructures(	IN		ibisGlobal_Pt	global_P,
						INOUT	ibisTOP_Pt		ibis_P )

{
ibisComponent_Pt	currentComponent_P;

copyGlobalDataToComponents( global_P, ibis_P );

copyComponentDataToModels( ibis_P );

for ( currentComponent_P = ibis_P->components_P;
	  currentComponent_P != NULL;
	  currentComponent_P = currentComponent_P->nextComponent_P )
	{
	if ( currentComponent_P->hasPinMapping )
		if ( copyPinMapToPins( currentComponent_P ) )
			return( 1 );
	}

for ( currentComponent_P = ibis_P->components_P;
	  currentComponent_P != NULL;
	  currentComponent_P = currentComponent_P->nextComponent_P )
	{
	if ( linkPinsToModels( currentComponent_P ) )
		return( 1 );
	}
return( 0 );
}
